/* file femgauss.c
  
Ce programme resout une EDP elliptique generale d'ordre 2
par elements finis P1 et factorisation de Gauss
La triangulation doit etre dans le fichier mesh.dta et les 
parametres et donnees de l'EDP dans param.dta. Ces deux fichiers
peuvent etre generes par freefem en utilisant l'option "saveall".
Resout  
   alpha u + (u1,u2)grad u- div(mat[rho] grad u) = f in , 
   u|_{ng!=0 && u0!=0} = u0,  (si u0=0 le changer en u0=eps<<1)
   (beta u + n^T mat[rho]grad u)|_{ng!=0} = g
avec toutes les donnees P^1 
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* derivees des fonctions de bases */
#define dwya(i,k) (-(q[me[k][next[i]]].x-q[me[k][next[i+1]]].x)/2)  
#define dwxa(i,k) ((q[me[k][next[i]]].y-q[me[k][next[i+1]]].y)/2) 	
#define dwy(i,k) (dwya(i,k)/area[k])    /* dw^i/dx sur T^k */
#define dwx(i,k) (dwxa(i,k)/area[k]) 		/* dw^i/y sur T^k */
#define sqr(x) ((x)*(x))
#define norm(x,y)(sqrt(sqr(x)+sqr(y)))
#define mmax(a,b)(a>b?a:b)
#define mmin(a,b)(a<b?a:b)
#define abss(a)(a >= 0 ? a : -(a))

#define penal 1.0e10

typedef struct{ float x,y;} rpoint;
typedef int triangle[3];
typedef struct{
 				int np, nt;	
 				rpoint *rp;		
	 			triangle *tr;	
 				int* ngt;		
 				int* ng;		
}triangulation;

/* prototypes */
int    initFEM(triangulation* t);
float  pdeian(float *, float *, float *, float *, float *, 
   float *, float *, float *, float *, float *, float *, 
   float *, float *, int, int);
float gaussband(float *, float *, int, long, int, float);
void rhsPDE(float*  fw, float*  f, float*  g);
int   buildarea();
void pdemat (float *a, float* alpha, float *rho11, float *rho12, 
  float *rho21,	float *rho22, float* u1, float* u2, float* beta);
int   readtriangulation( triangulation *, char * );
int   savefct(float *,  int , char *);
int   readparam(triangulation* , char *);

/* variables globales */
 triangulation t;

 int next[5]={1,2,0,1,2};  /* i+1 mod 3 */
 char* pathT = "mesh.dta"; /* fichier maillage */
 char* pathP = "param.dta";/* fichier parametre coef EDP */
 char* pathU = "sol.dta";  /* fichier resultat */
   float* area; 	/* aires des triangles */
   float* a;    	/* matrice du systeme */
   float* u; 			/* solution */
   float* f; 			/* sources */
   float* g; 			/* Neumann */
   float* u0; 		/* Dirichlet */
   float* beta; 	/* Coef ordre 0 dans Fourier */
   float* alpha; 	/* Coef ordre 0 EDP */
   float* rho11; 	/* Matrice de viscosite' */
   float* rho12; 
   float* rho21; 
   float* rho22;
   float* u1; 		/* vitesse de convection */
   float* u2; 			
   int factorize = 1; /* pour factoriser la matrice */
/* Copie de la triangulation */
 		int ns, nt ; 
 		rpoint* q;
		triangle* me;
 		int* ng;
 		int* ngt;
 		int bdth;  
 
void rhsPDE( float* fw, float* f, float* g)
/* ------------------------------------------
  Calcul le second membre du systeme lineaire
  fw(j)=\int_\Omega f w^j + \int_{ng<>0} g w^j+penal|_{ng<0} u0^j 
  AUTRE INPUT next,area,q,me,ng,nt,penal
*/
{ int j,k, ir, ir1, ir2, meirk, meirkn ;
	float x1,x2, aux;
	
	for(j=0;j<ns;j++) fw[j] = 0 ;
  for(k=0;k<nt;k++)
    for(ir=0;ir<=2;ir++)
    {							
			ir1 = next[ir]; ir2 = next[ir1];
			meirk = me[k][ir]; meirkn = me[k][ir1];
	    fw[meirk]+=(2*f[meirk]+ f[meirkn]+ f[me[k][ir2]])*area[k]/12;
		}	      
    for(k=0;k<nt;k++)
    	for(ir=0;ir<=2;ir++)
    	{							
				ir1 = next[ir]; ir2 = next[ir1];
				meirk = me[k][ir]; meirkn = me[k][ir1];
        if ((ng[meirk] != 0)&&(ng[meirkn] != 0))
        {
          aux = norm(q[meirk].x-q[meirkn].x,q[meirk].y-q[meirkn].y)/6;
         	x1 = g[meirk] * aux;
				  x2 = g[meirkn] * aux;
				  fw[meirk] += 2*x1 + x2;
         	fw[meirkn] += x1 + 2 * x2;
        }
      }
}

int buildarea(void)
/*-------------------------------------------
  Calcul les aires: INPUT la triangulation ns,nt,me,q
  RETOURNE 1 si une aire est <0, 0 sinon
*/
{ int i,k,ir, err = 0;
	float qq[2][3];
	for( k=0;k<nt; k++)
	{
		for(ir=0;ir<=2;ir++)
		{
      i=me[k][ir]	;	 qq[0][ir]=q[i].x;		 qq[1][ir]=q[i].y ;
 		}
    area[k]=( (qq[0][1]-qq[0][0])*(qq[1][2]-qq[1][0])
              - (qq[1][1]-qq[1][0])*(qq[0][2]-qq[0][0]) )/2 ;
    err = (area[k] < 0);
 	}
 	return err;
}
 
void pdemat (float *a, float* alpha, float *rho11, float *rho12, 
  float *rho21,	float *rho22, float* u1, float* u2, float* beta)
/* ------ Calcul la matrice --------------------------------- */
{
	int k,i,j,mejk, meik, k0, k1, k2;
	float rhomean[2][2], alphamean, isii, x1, x2, x3, x4, aux ;
	long ai, nsl=ns, nsbdt = (2*bdth+1)*nsl;

	for(ai=0;ai< nsbdt;ai++) a[ai] = 0.0F ;
  for(k=0;k<nt;k++)
		for(i=0;i<=2;i++)
		{
		  meik = me[k][i] ; 
		  k0 = meik; k1=me[k][next[i]]; k2=me[k][next[i+1]] ;
      x1=rho11[k0]; x2=rho11[k1]; x3= rho11[k2] ;
			rhomean[0][0] = (x1+x2+x3) / 3 ;
       		x1=rho12[k0]; x2=rho12[k1]; x3= rho12[k2] ;
			rhomean[0][1] = (x1+x2+x3) / 3 ;
       		x1=rho21[k0]; x2=rho21[k1]; x3= rho21[k2] ;
			rhomean[1][0] = (x1+x2+x3) / 3 ;
       		x1=rho22[k0]; x2=rho22[k1]; x3= rho22[k2] ;
			rhomean[1][1] = (x1+x2+x3) / 3 ;
       		x1 = alpha[k0]; x2 = alpha[k1]; x3 = alpha[k2];
			alphamean = ( x1 +x2 +x3 ) / 3 ;
			for(j=0;j<=2;j++)
			{
       	mejk = me[k][j];
       	isii = i==j ? 1.0F/6.0F : 1.0F / 12.0F ;
       	ai = nsl*(meik-mejk+bdth)+mejk;
				aux = dwxa(i,k) * dwx(j,k);
				x1 =    rhomean[0][0] * aux;
				aux = dwya(i,k) * dwx(j,k);
				x2 = rhomean[1][0] * aux;
				aux = dwxa(i,k) * dwy(j,k);
				x3 = rhomean[0][1] * aux;
				aux =  dwya(i,k) * dwy(j,k);
				x4 =  rhomean[1][1] *aux;
				a[ai]  += x1 + x2 + x3 + x4;
				x1 = u1[k0]; x2 = u1[k1];  	x3 = u1[k2] ;
				a[ai]  +=  (2*x1  + x2 + x3) * dwxa(j,k) / 12;
				x1 = u2[k0]; x2 = u2[k1];  x3 = u2[k2] ;
				a[ai]  +=  (2*x1  + x2 + x3) * dwya(j,k) / 12 
					   +  alphamean * area[k] * isii;
      	if((ng[meik] != 0)&&(ng[mejk] != 0)&&(meik < mejk))
      	{
      	  k0 = meik; k1 = mejk;
      		x1 = beta[k0]; x2 = beta[k1];
					x1 = (x1 + x2) * 
					  norm(q[meik].x-q[mejk].x,q[meik].y-q[mejk].y)/2;
					a[ai] += x1/6;
					ai = nsl*bdth+mejk; a[ai] += x1/3;
					ai = nsl*bdth+meik; a[ai] += x1/3;
      	}	
     }
  }
}
 
/*----------------------------------------------------*/
float gaussband (float *a, float *x, int n, long bdthl, 
                   int first, float eps)
/*----------------------------------------------------*/
/* Factorise (first=1) et/ou resout Ay = x avec resultat en x */
/* LU est stocke' dans A ; retourne le plus petit pivot  */
/*  Tout pivot < eps est mis a eps */
/*  a[i][j] est stocje' dans a[i-j+bdthl][j]=a[n*(i-j+bdthl)+j)*/
/*  ou -bdwth <= i-j <= bdthl */
{
	int i,j,k;	 
	float s, s1, smin = 1e9;
	
	if (first)			/* factorization */
	  for (i=0;i<n;i++) 
		{
			for(j=mmax(i-bdthl,0);j<=i;j++)
			{
			  s=0; for (k=mmax(i-bdthl,0); k<j;k++) 
							s += a[n*(i-k+bdthl)+k]*a[n*(k-j+bdthl)+j] ;
			  a[n*(i-j+bdthl)+j] -= s ;
			}
			for(j=i+1;j<=mmin(n-1,i+bdthl);j++)
			{
				s=0; for (k=mmax(j-bdthl,0);k<i;k++) 
					     s += a[n*(i-k+bdthl)+k]*a[n*(k-j+bdthl)+j] ;
				s1 = a[n*bdthl+i];
				if(abss(s1) < smin) smin=abss(s1);
				if(abss(s1) < eps) s1 = eps;
				a[n*(i-j+bdthl)+j] = ( a[n*(i-j+bdthl)+j] - s)/s1;
			}
		}
	 for (i=0;i<n;i++)							/*  resolution */
		{	
			s=0; for (k=mmax(i-bdthl,0);k<i;k++)
			       s += a[n*(i-k+bdthl)+k] * x[k];
			x[i] = (x[i] - s ) / a[n*bdthl+i] ;
		}
	 for (i=n-1;i>=0;i--)
		{
			s=0; for (k=i+1; k<=mmin(n-1,i+bdthl);k++)
			       s += a[n*(i-k+bdthl)+k] * x[k];
			x[i] -= s ;
		}
	return smin;	
}

int initFEM(triangulation* t)
{
	int i,k, baux;
	long matsize;
  ns = t->np; 	nt = t->nt;
  q =t->rp; 		me = t->tr;
  ng = t->ng; 	ngt =t->ngt;
	bdth = 0;
	for( k=0;k<nt;k++)
		for(i=0;i<=2;i++)
		{
			baux = abss((me[k][i] - me[k][next[i]])); 
			bdth = (bdth > baux)? bdth : baux;
		}
 	area = (float*)calloc(nt, sizeof(float));
	matsize = (long)ns*(2*(long)bdth+1);
	a = (float*)calloc(matsize, sizeof(float));
	return buildarea();
}
	

	
/* ============I/O fonctions ==================== */
int readtriangulation( triangulation *t, char *path )
{
  int i, j, ns, nt;
  FILE *data;

  if(data=fopen(path, "r"), !data) return 2;
  fscanf( data, "%d	%d \n", &t->np, &t->nt );  
  ns = t->np; nt = t->nt;
  t->rp = (rpoint*)calloc(ns, sizeof(rpoint));
  t->tr = (triangle*)calloc(nt, sizeof(triangle));
  t->ng = (int*)calloc(ns, sizeof(int));
  t->ngt = (int*)calloc(nt, sizeof(int));
  for( i=0; i<t->np; i++ )
    fscanf( data, "%f	%f	%d\n", &t->rp[i].x, &t->rp[i].y, 
                   &t->ng[i] ); 
  for( i=0; i<t->nt; i++ )
  {
    for( j=0; j<3; j++ ) 
    {
    	fscanf( data, "%d	", &t->tr[i][j] ); 
    	t->tr[i][j]--; /* fortran standard */
    }
    fscanf( data, "%d\n",&t->ngt[i] ); 
  }
  fclose(data);
  return 0;
}


int savefct(float *f,  int ns, char *path)
{
  int i;
  FILE *data;

  if(data=fopen(path, "w"), !data) return 2;
  fprintf(data, "%d\n", ns);
  for(i=0; i<ns; i++) fprintf(data, "%f\n", f[i]);
  fclose(data);
  return 0;
}

int readparam(triangulation* t, char *path)
{
  int i, ns = t->np;
  FILE *data;

  if(data=fopen(path, "r"), !data) return 2;
  fscanf(data, "%d	%d", &ns,&i); /* i=1 pour eq scalaire */
  if(ns != t->np) return 3;
  beta = (float*)calloc(ns, sizeof(float));
  g = (float*)calloc(ns, sizeof(float));
  u0 = (float*)calloc(ns, sizeof(float));
  f = (float*)calloc(ns, sizeof(float));
  alpha = (float*)calloc(ns, sizeof(float));
  u1 = (float*)calloc(ns, sizeof(float));
  u2 = (float*)calloc(ns, sizeof(float));
  rho11 = (float*)calloc(ns, sizeof(float));
  rho22 = (float*)calloc(ns, sizeof(float));
  rho12 = (float*)calloc(ns, sizeof(float));
  rho21 = (float*)calloc(ns, sizeof(float));
  for(i=0; i<ns; i++) 
  fscanf(data, "%f	%f	%f	%f	%f	%f	%f	%f %f %f %f",	
  						&f[i],
  						&g[i],
  						&u0[i],
  						&alpha[i],
  						&beta[i], 
  						&u1[i],
  						&u2[i],
  						&rho11[i],
  						&rho12[i],
  						&rho21[i],
  						&rho22[i]);
  fclose(data);
  u = (float*)calloc(ns, sizeof(float)); 
  return (u==0);
}


void main( )
/*-----------
Resout  alpha u + (u1,u2)grad u- div(mat[rho] grad u) = f in , 
         u|_{ng!=0 && u0!=0} = u0, 
         (beta u + n^T mat[rho]grad u)|_{ng!=0} = g
avec f dans P^1 
*/
{
   int i; long ai, bdthl;
   float err;
    
  if(readtriangulation( &t, pathT))
  {		printf("erreur lecture triangulation\n"); exit(0);}
  initFEM(&t); bdthl =bdth;
  if(readparam(&t,pathP) != 0) 
    {		printf("erreur lecture des parametres\n"); exit(0);}
	if(factorize) 
	{
		pdemat(a,alpha,rho11,rho12,rho21,rho22, u1, u2, beta);
		for(i=0;i<ns;i++)
			if(u0[i] !=0 ){ ai = ns*bdthl+i; a[ai] += penal;}
	}
	rhsPDE(u, f, g);
	for(i=0;i<ns;i++)
		if(u0[i] !=0)	u[i] += u0[i]*penal;
	err = gaussband(a,u, ns, bdthl, factorize,1.0/penal);
	printf("smallest pivot = %f \n",err);
	savefct(u,ns, pathU); 
}

/* TEST DE femgauss.c */
/* -------------------*/
/* 
1. Choisir une solution, par exemple ue = sin(x+y) 
Choisir les parametres de l'EDP pour retrouver cette solution.

2. Resoudre avec freefem. Ici le programme freefem sera:
   
changewait;
ns:=40; 
border(1,0,2*pi,2*ns) begin x:=     3*cos(t); y:= 2*sin(t); end;
border(2,0,2*pi,ns) begin x:= cos(-t); y:= sin(-t); end;
buildmesh(ns*ns);
savemesh('mesh.dta');

ue= sin(x+y); 
p = ue; nx = -x; ny =- y; dxue = cos(x+y);
c = 0.2; a1 = y; a2 = x; b=1;
nu = 1; nu11 = 1; nu22 = 2; nu21 =0.3; nu12 =0.4;
dnuue=dxue*(nu*(nx+ny) + 
(nu11 + nu12)*nx + (nu21+ nu22)*ny);
g = ue*c+dnuue;
f = b*ue+dxue*(a1+a2) +ue*(2*nu+nu11+nu12+nu21+nu22);

solve(u) begin
onbdy(1) u = p;
onbdy(2) id(u)*c + dnu(u) = g;
pde(u) id(u)*b + dx(u)*a1 + dy(u)*a2
-laplace(u)*nu - dxx(u)*nu11 - dxy(u)*nu12 
    - dyx(u)*nu21 - dyy(u)*nu22 =f;
end;

plot(u-ue);

3. On verifie que u-ue est petit.

4. Pour generer les donnees pour femgauss.c utiliser le meme
programme pour freefem ou "solve" est remplace' par

    saveall('param.dta',u)
    
5. La solution calcule' par femgauus.c se trouve dans 'sol.dta'
verifier que c'est la bonne en executant le programme freefem
suivant

loadmesh('mesh.dta');
load('sol.dta',u);
plot(u);
plot(sin(x+y));
plot(u-sin(x+y));

*/
