#include <iostream.h>
#include <math.h>

#include "aarray.h"
#include "matrix.h"
#include "linsolve.h"

#define mmax(a,b)(a>b?a:b)
#define mmin(a,b)(a<b?a:b)
#define abss(a)(a >= 0 ? a : -(a))

float gaussband (Bandmatrix& a, Vector& x, int first)
/*----------------------------------------------------
 Factorise (first=1) and/or solve 	Ay = x  with result in x 
 LU is stored in A ; returns the value of the smallest pivot  
  all pivots less than eps are put to eps 
  a[i][j] is stored in a[i-j+bdthl][j]=a[n*(i-j+bdthl)+j] 
  where -bdwth <= i-j <= bdthl 
*/
{
	int i,j,k;	 
	float s, s1, eps = 1e-15;
	float smin = 1e9;
	int n = a.n;
	int bdthl = a.bdth;
	
	if (first)			
	    for (i=0;i<n;i++) 
		{
			for(j=mmax(i-bdthl,0);j<=i;j++)
			{
				s=0; for (k=mmax(i-bdthl,0); k<j;k++)  s += a(i,k)*a(k,j) ;
				a(i,j) -= s ;
			}
			for(j=i+1;j<=mmin(n-1,i+bdthl);j++)
			{
				s=0; for (k=mmax(j-bdthl,0);k<i;k++)  s += a(i,k)*a(k,j) ;
				s1 = a(i,i);
				if(abss(s1) < smin) smin=abss(s1);
				if(abss(s1) < eps) s1 = eps;
				a(i,j) = ( a(i,j) - s)/s1;
			}
		}
	 for (i=0;i<n;i++)							
		{	
			s=0; for (k=mmax(i-bdthl,0);k<i;k++)  s += a(i,k) * x[k];
			x[i] = (x[i] - s) /  a(i,i) ;
		}
	 for (i=n-1;i>=0;i--)
		{
			s=0; for (k=i+1; k<=mmin(n-1,i+bdthl);k++)  s += a(i,k) * x[k];
			x[i] -= s ;
		}
	return smin;	
}

float gradconj(Bandmatrix& a, Vector& x, Vector& b, float epsilon, int kmax)
/*----------------------------------------------------
Solves ax=b with result in x by conjugate gradient method
preconditionned by the diagonal of a.
-----------------------------------------------------*/
{
  int i,j,k;
  int n = x.size;
  float gcg,gamma,hah, ro, eps;
  Vector h(n), ah(n), cg(n), g(n);
  for(k=0;k<kmax;k++)
  {
	if(k==0)
	{
     	h = a * x;
     	cg = b - h;
     	g = cg;
     	for(i=0;i<n;i++) g[i] /= a(i,i);
        gcg = g * cg;
        h = g;
     }
     ah = a * h;
     hah = h * ah;
     ro=gcg / hah;
     cg -= ro * ah;
     for(i=0;i<n;i++) ah[i] /= a(i,i);
     x += ro * h;
     g -= ro * ah;
     gamma=gcg;
     gcg = g * cg;
     cout <<"at iteration: "<<k<<"  gcg= "<<gcg<<endl;
     if(k==0) epsilon *= gcg;
     if (gcg < epsilon) return gcg;
     gamma=gcg/gamma;
     h = g + gamma * h;
	}
	cout<< "Stopped before convergence"<<endl;
	return gcg;
}


 int gmrs (Bandmatrix& a, Vector& x,  Vector& b, float eps,int nkr,int nit)
/*--------------------------------------------------------------
   Solves ax = b by GMRES with nit iterations (reinitializations)
   and nkr Krilov vectors.
 */
{
	int i,j, nk, i0,im,it;
	float tem=1, res, rec,zero, ccos,ssin;
	long ns = x.size;
	Matrix v(ns,nkr); 			// Krilov vectors
	Matrix h(nkr+1,nkr);		// Heisenberg maatrix of coefficients
	Vector r(nkr+1),v0(ns);		// linear combination coef of Krilov vectors and working array
	rec=zero = 1.e-8;
	
//initialisation
     v0 = b - a * x;
     res =sqrt( v0 * v0);
     cout<<"initial residual = "<<res<<endl;
     if(res<rec) return 0; // nothing to do
     rec = (rec<res*eps) ? res*eps : rec;
     cout<<"Stopping rule:"<<rec<<endl;

// iterations
     for( it=0;it<nit;it++)
	  {
		nk = nkr;
		
//...orthogonalisation of Arnoldi
		v0 /= res;
		r = 0; r[0] = res;
		
		for( j=0;j<nkr;j++)
		{											  
		  v(j) = v0;
		  v0 = a * v0;

		  for(i=0;i<=j;i++)
		  {
			h(i,j) = v0 * v(i);										   
			v0 -= h(i,j) * v(i);											 
		  }
		  tem = sqrt(v0 * v0);										   
		  h(j+1,j) = tem;
		  if(tem<rec) 
		  {
			nk = j+1;
			cout<<"tem="<<tem<<" nk="<<nk<<endl;
			goto l5;
		  }
		  v0 /= tem;
		}
																	 
//...Triangularisation
l5:		  	for(i=0;i<nk;i++)
		  	{														
			  im = i+1; 
			  tem = 1/sqrt(h(i,i)*h(i,i) + h(im,i)*h(im,i));													   
			  ccos = h(i,i) * tem;
			  ssin = - h(im,i) * tem;									
			  for(j=i; j<nk;j++)
			  {
			  	tem = h(i,j);
			  	h(i,j) = ccos * tem - ssin * h(im,j);
			  	h(im,j) =  ssin * tem + ccos * h(im,j);
			  }
			  r[im] = ssin * r[i]; r[i] *= ccos; 			  
		  	}

// Solution of linear system
	  for( i=nk-1;i>=0;i--)												  
	   {     	  
	   	  r[i]  /= h(i,i);						  
		  for( i0=i-1;i0>=0;i0--)
		  	r[i0] -= h(i0,i)* r[i]; 
		} 
	   for( i=0; i<nk;i++) x += r[i]*v(i); 

// New residual and stopping tests
		v0 = b - a * x;
     	res = sqrt(v0 * v0);
     	cout<<"at it="<<it<<" residu ="<<res<<endl;
     	if(res<rec) return 1; 
		else if (it==nit-1) cout<<"Stopped before convergence"<<endl;
	}
return -1;
}	

void main() 
{	// Test program
	long ij,i,j,nv = 100, bdth=10;
	int nkrilov, niter;
	Vector b(nv);
	Vector x(nv);
	Bandmatrix a(nv, bdth);
	
	for(i=0;i<nv;i++)
	  {
		for(j=mmax(i-bdth,0);j<=mmin(nv-1,i+bdth);j++) 
		{	
			ij = nv*(i-j + bdth) + j;
			if((i-j<4)||(j-i)<4)
			 if(i<j) a(i,j) = -1; else a(i,j) = -1;
			else a(i,j)=0;
		} 
		a(i,i) = 12;
	  }
	x = 1;
	b = a * x; 
	x = 0;
  cout << gradconj(a,x,b,1.e-5,50)<< endl;
  for(i=0;i<10;i++) cout<<x[i]<<endl;
  x = 0;
  cout << gmrs ( a, x, b, 1.e-5,nkrilov=10,niter=40) << endl;
  for(i=0;i<10;i++) cout<<x[i]<<endl;
  cout << "Pivot= " << gaussband ( a,  b, 1) << endl;
  for(i=0;i<10;i++) cout<<b[i]<<endl;
}