#include <iostream.h>
#include <math.h>

#include "aarray.h"
#include "matrix.h"

#define mmax(a,b)(a>b?a:b)
#define mmin(a,b)(a<b?a:b)
#define abss(a)(a >= 0 ? a : -(a))

const float eps=1.e-10; // precision for pivot in gausband


// 							IMPLEMENTATION
  Vector::Vector ( const Vector& v ) 
  { size=v.size; 
  	cc = new base[size]; 
  	for (int i=0; i< size; i++ ) cc[i] = v.cc[i];
  }

  const Vector& Vector::operator = ( const Vector& v)
  {	if ( this != &v )	{	delete [] cc; size = v.size; cc = new base[v.size];
      						for (int i=0; i<size;i++) cc[i] = v.cc[i];
    					}     return *this;
  }

  const Vector& Vector::operator = ( const float x)
  {	
  	for (int i=0; i<size;i++) cc[i] = x;
    return *this;
  }

float Vector::norm () const
{	float res = 0; for ( int i = 0; i<size; i++ ) res += cc[i]*cc[i]; return sqrt(res);}

float operator * (const Vector& v1, const Vector& v2 )
{	float res = 0; 
	myassert(v1.size == v2.size);
	for ( int i = 0; i<v1.size; i++ ) res += v1.cc[i]*v2.cc[i]; 
	return res;
}

Vector operator * ( const base scal, const Vector& v )
{  Vector v2(v.size);  
	for (int i=0; i<v.size ;i++)  v2[i] = v[i]*scal;  
	return v2;
}


Vector operator -(const Vector& v1, const Vector& v2 )
{	  
	myassert(v1.size == v2.size);
	Vector v(v1.size);
	for (int i=0; i<v1.size ;i++)  v[i] = v1[i] - v2[i]; 
	return v;
}

Vector operator +(const Vector& v1, const Vector& v2 )
{	  
	myassert(v1.size == v2.size);
	Vector v(v1.size);
	for (int i=0; i<v1.size ;i++)  v[i] = v1[i] + v2[i]; 
	return v;
}

Vector& Vector::operator /=(const float x)
{
		for(int i=0;i<size;i++) cc[i] /= x; 
		return *this;
}

Vector& Vector::operator +=(const Vector& x)
{
		for(int i=0;i<size;i++) cc[i] += x[i]; 
		return *this;
}

Vector& Vector::operator -=(const Vector& x)
{
		for(int i=0;i<size;i++) cc[i] -= x[i]; 
		return *this;
}

ostream& operator << ( ostream& os, Vector& c )
{	 for(int i=0; i<c.size;i++) os << c[i] << endl; return os;}


// IMPLEMENATION CLASS Matrix
    
Matrix::Matrix ( int mm, int nn )
  { n=nn; m= mm; notfact=1;
  	ccc = new Vector[nn];
  	for(int i =0;i<nn;i++) 
  	{
  		ccc[i].size = mm; ccc[i].cc = new base[mm];
  	}
  }

Matrix::Matrix ( Matrix& a )
  { n=a.n; m= a.m; notfact=a.notfact;
  	ccc = new Vector[n];
  	for(int i =0;i<n;i++) 
  	{
  		ccc[i].size = m;
  		 ccc[i].cc = new base[m]; 
  		ccc[i] = a.ccc[i];
  	}
  }

Vector operator * (const Matrix& a, const Vector& x)
{
	myassert(a.n==x.size);
	Vector b(a.n);
	for(int i=0;i<a.m;i++)
	{
		float bcc = 0; for(int j=0;j<a.n;j++) bcc += a.ccc[j].cc[i] * x.cc[j];
		b.cc[i] = bcc;
	}
	return b;
}


// IMPLEMENATION CLASS Bandmatrix
Vector operator * (const Bandmatrix& a, const Vector& x)
{	Vector b(a.n); float s; long ik;
			myassert(x.size==a.n);
			for(int i=0;i<a.n;i++)
			{
				s=0; for (int k=mmax(i-a.bdth,0); k<=mmin(a.n-1,i+a.bdth);k++)
				  	 {	ik = (long)a.n*(i-k + a.bdth) + k;
				  		s += a.cc[ik]*x.cc[k] ;
				  	 }
				b.cc[i] = s ;
			}
	return b;
} 

ostream& operator << (ostream& os, Bandmatrix& a)
{
	int i,j;
	for(i=0; i<a.n;i++)
	{	for(j=0;j<a.n; j++)
			if( abss(i-j) <= a.bdth ) os << a(i,j)<<"	"; else os <<"	0	";
		os << endl;
	}
    return os;
}

float gaussband (Bandmatrix& a, A<float>& x, int first)
/*----------------------------------------------------
 Factorise (first=1) and/or solve 	Ay = x  with result in x 
 LU is stored in A ; returns the value of the smallest pivot  
  all pivots less than eps are put to eps 
  a[i][j] is stored in a[i-j+bdthl][j]=a[n*(i-j+bdthl)+j] 
  where -bdwth <= i-j <= bdthl 
*/
{
	int i,j,k;	 
	float s, s1;
	float smin = 1e9;
	int n = a.n;
	int bdthl = a.bdth;
	
	if (first)			
	    for (i=0;i<n;i++) 
		{
			for(j=mmax(i-bdthl,0);j<=i;j++)
			{
				s=0; for (k=mmax(i-bdthl,0); k<j;k++)  s += a(i,k)*a(k,j) ;
				a(i,j) -= s ;
			}
			for(j=i+1;j<=mmin(n-1,i+bdthl);j++)
			{
				s=0; for (k=mmax(j-bdthl,0);k<i;k++)  s += a(i,k)*a(k,j) ;
				s1 = a(i,i);
				if(abss(s1) < smin) smin=abss(s1);
				if(abss(s1) < eps) s1 = eps;
				a(i,j) = ( a(i,j) - s)/s1;
			}
		}
	 for (i=0;i<n;i++)							
		{	
			s=0; for (k=mmax(i-bdthl,0);k<i;k++)  s += a(i,k) * x[k];
			x[i] = (x[i] - s) /  a(i,i) ;
		}
	 for (i=n-1;i>=0;i--)
		{
			s=0; for (k=i+1; k<=mmin(n-1,i+bdthl);k++)  s += a(i,k) * x[k];
			x[i] -= s ;
		}
	return smin;	
}
