C  fichier: minEe.for
C
C---------------------------------------------------------
        FUNCTION E(q,me,phi,f,ns,nt)
C---------------------------------------------------------
C
C Calcule l'energie E. Cette fonction peut etre utile pour 
C verifier que l'energie decroit effectivement au cours 
C des iterations de la methode employee pour la minimiser.
C
C Si le tableau f en entree de ce sous-programme est 
C identiquement nul, cette fonction calcule alors
C $\int_\Omega{1\over 2}|\nabla \phi|^2$, pour tout tableau phi(ns), ce qui
C fournit une autre utilite a cette fonction.
C
C     tableaux:
C     --------- 
C     maillage:
C       q(2,ns):coordonnees des sommets du maillage,
C       me(3,nt):numerotation des sommets des triangles,
C
C     donnees:
C       f(ns):second membre du Laplacien,
C       phi(ns):au depart, conditions aux limites de 
C           Dirichlet.
C
C     scalaires:
C     ---------
C       fk:valeur de f au centre de gravite du triangle 
C         d'indice k, ou de maniere equivalente, f etant
C         suppose affine sur chaque triangle, le tiers de 
C         la somme des valeurs prises par f en chacun 
C         des 3 sommets du triangle d'indice k.
C       aire2Tk:double de l'aire du triangle d'indice k
C
C
      DIMENSION phi(ns),f(ns),me(3,nt),q(2,ns)
C
      E = 0.
      DO 1 k = 1,nt
        i1 = me(1,k)
        i2 = me(2,k)
        i3 = me(3,k)
        xi1=q(1,i1)                                              
        xi2=q(1,i2)                                              
        xi3=q(1,i3)                                              
        yi1=q(2,i1)                                              
        yi2=q(2,i2)                                              
        yi3=q(2,i3)
        phii1=phi(i1)
        phii2=phi(i2)
        phii3=phi(i3)
C
        fk=(f(i1)+f(i2)+f(i3))/3.                                              
        aire2Tk = deter(xi1,xi2,xi3,yi1,yi2,yi3) 
        gradFx = deter(phii1,phii2,phii3,yi1,yi2,yi3)
     >                          /aire2Tk
        gradFy = deter(xi1,xi2,xi3,phii1,phii2,phii3)
     >                          /aire2Tk
C
        E = E + 0.5*abs(aire2Tk)*(
     >              (gradFx*gradFx + gradFy*gradFy)/2.
     >             -fk*(phii1+phii2+phii3)/3. )
1     CONTINUE
C
      RETURN
      END

