C Fichier: Arete.for
C
C----------------------------------------------------------
      SUBROUTINE calArete(ns,nt,me,list,listTete,iTGauche
     >           ,iTDroit,infAr,isupAr,nArT,na)
C----------------------------------------------------------
C Ce sous-programme numerote les aretes et construit
C les tableaux suivants: iTGauche,iTDroit
C infAr,isupAr,nArT.
C Il calcule le nombre total na d'aretes.
C Cette construction utilise les tableaux auxiliaires
C d'adresses: list et listTete.
C
C     entree: triangulation
C     -------
C     ns: nombre de sommets,
C     nt: nombre de triangles,
C     me(3,nt): numerotation locales des sommets 
C       de chaque triangle (me(il,k)= indice du 
C       il eme sommet du triangle d'indice k);
C     
C     N.B. Les triangles sont supposes orientes
C          dans le sens positif. Le tableau ng 
C          n'est pas necessaire ici.
C
C     sortie:
C     -------
C     na: nombre total d"aretes
C     infAr(na), isupAr(na): indices des points 
C       extremaux de l"arete avec la convention:
C       infAr(ia) $<$ isupAr(ia);
C     iTGauche(na), iTDroit(na): indices des 
C       triangles situes a gauche et a droite de
C       chaque arete, en convenant d"orienter 
C       l"arete d'indice ia du point d"indice
C       infAr(ia) vers le point d'indice isupAr(ia);
C     nArT(3,nt): indice des 3 aretes de chaque 
C       triangle, avec la convention:
C       l"arete d"indice nArT(ial,k) est opposee, 
C       dans le triangle d"indice k, au point 
C       d"indice me(ial,k).
C
C     tableaux auxiliaires:
C     ---------------------
C     list(*): contient tous les indices d'"aretes
C       demarrant d"un point donne
C     listTete(*): pour chaque sommet d"indice i, 
C       listTete(i) contient l"indice de l"arete 
C       d"indice le plus eleve demarrant
C       du sommet d"indice i; l"arete suivante a
C       pour indice list(listTete(i),...
C
C     N.B.1 Les valeurs de iTGauche(ia) et iTDroit(ia) 
C          sont construites a 2 moments differents:
C          l'une au moment de la construction de l"arete, 
C          l'autre au moment ou on rencontre cette 
C          arete pour la deuxieme fois.
C
C     N.B.2 Par la formule d"Euler, on a:
C          na $<$ ns+nt
C 
C
      DIMENSION me(3,nt)
      DIMENSION infAr(*),isupAr(*),iTGauche(*),iTDroit(*)
      DIMENSION nArT(3,nt)
      DIMENSION list(*),listTete(*)
C
      na=0
      DO 4 ia=1,ns+nt
4     iTDroit(ia)=0
      DO 1 i=1,ns
1     listTete(i)=0
      DO 2 k=1,nt
        DO 2 il=1,3
         ilp=mod(il,3)+1
         i=me(ilp,k)
         j=me(mod(ilp,3)+1,k)
         ipetit=min(i,j)
         igrand=max(i,j)
         ia=listTete(ipetit)
C
C On parcourt le tableau list pour savoir si l"arete
C a deja ete creee.
C
3        IF( ia.NE.0) THEN
          IF( (igrand.EQ.isupAr(ia) ).AND.
     >     ( ipetit.EQ.infAr(ia) ) )THEN
            nArT(il,k)= ia
            IF (ipetit.EQ.i) THEN
              iTGauche(ia) = k
            ELSE
              iTDroit(ia)=k
            ENDIF
            GOTO 2
          ENDIF
          ia=list(ia)
          GOTO 3
         ENDIF
C
C L"arete est nouvelle
C
       na=na+1
       infAr(na)=ipetit
       isupAr(na)=igrand
       list(na)=listTete(ipetit)
       listTete(ipetit)=na
       IF (ipetit.EQ.i) THEN
         iTGauche(na)=k
       ELSE
         iTDroit(na) = k
       ENDIF
       nArT(il,k)=na
2      CONTINUE
       END
C
C
       PROGRAM listArete
C      -----------------
C
C Ce programme a pour but de tester le sous-
C programme precedent dans le cas de la 
C triangulation minuscule. La triangulation
C est supposee etre stockee dans le fichier
C TinyTri.dta cree au chapitre precedent 
C (paragraphe 5.2); on la lit donc par 
C l"intermediaire du sous-programme litT
C stocke dans le fichier utilP1.for.
C
C
      PARAMETER(nsMax=8,ntMax=8,naMax=nsMax+ntMax)
      INTEGER list(3*ntMax), listTete(nsMax)
      INTEGER infAr(naMax),isupAr(naMax)
      INTEGER iTGauche(naMax),iTDroit(naMax)
      INTEGER nArT(3,ntMax)
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
C
      CALL litT('TinyTri.dta',ns,nt,q,me,ng,ngt)
      CALL  calArete(ns,nt,me,ng,list,listTete
     >    ,iTGauche,iTDroit,infAr,isupAr,nArT,na)
       Do 16 k=1,nt
16      write(*,*)(nArT(ial,k),ial=1,3)
       write(*,*)
       write(*,*) list
       write(*,*)listTete
       write(*,*)
       DO 15 ia=1,na
15      write(*,*)ia,infAr(ia),isupAr(ia)
     >          ,iTDroit(ia), iTGauche(ia)
       PAUSE
       END
C
C
C***************************************
C
C      BIBLIOTHEQUE
C
C***************************************
C
C
       INCLUDE 'utilP1.for'
C
C
