C Fichier: BCholes.for
C
C---------------------------------------------------------
      SUBROUTINE BCholeski(n,a,nBand)
C---------------------------------------------------------
C Ce sous-programme calcule la factorisee de Choleski L
C d'une matrice symetrique definie positive A de taille
C n*n, dont seule la partie triangulaire inferieure est 
C stockee sous forme bande dans le tableau a de taille 
C nMat=(n-1)*nBand+n.
C On suppose connue la largeur de bande nBand. La 
C factorisee de Choleski L a la meme structure bande
C que la matrice A.
C
C La factorisation se fait colonne par colonne, en  
C commencant par le calcul de l'element diagonal. La 
C matrice L est stockee a la place de A dans le tableau a.
C
C
C     entree:
C     -------
C     largeur de bande nBand de la matrice A(n,n),i.e.
C       $nBand = max\{|i-j|: i>j, A(i,j) \ne 0\}$;
C     tableau a(*) de stockage de la matrice A: A(i,j)
C       est stocke dans a((i-1)*nBand +j).
C
C
C     N.B. il faut tenir compte des contraintes sur les
C       indices imposees par le stockage Morse:
C       $max(1,i-nBand) \le j \le i \le min(ns,j+nBand)$
C
C
C     sortie:
C     -------
C      le tableau a(*) contient les elements de la matrice
C        L, i.e. a((i-1)*nBand +j) contient maintenant 
C        l'element L(i,j).
C
C
      DIMENSION a(*)
C
      DO 1 j=1,n
        sum=0.0
        j1=(j-1)*nBand
        kMin=max(1,j-nBand)
        DO 2 k=kMin,j-1
          jk=j1+k
          sum=sum+a(jk)**2
2       CONTINUE
         jj=j1+j
         a(jj)=sqrt(a(jj)-sum)
C
        iMax=min(n,j+nBand)
        DO 4 i=j+1,iMax
          sum=0.0
          i1=(i-1)*nBand
          kMin=max(1,i-nBand)
          DO 3 k=kMin,j-1
            ik=i1+k
            jk=j1+k
            sum=sum+a(ik)*a(jk)
3         CONTINUE
        ij=i1+j
        a(ij)=(a(ij)-sum)/a(jj)
4       CONTINUE
1     CONTINUE
      END
C
C
C---------------------------------------------------------
      SUBROUTINE BSolvSys(n,a,g,nBand)
C---------------------------------------------------------
C Ce sous-programme calcule la solution x du systeme 
C $Ax=g$, ou A est une matrice symetrique definie positive 
C de taille n*n factorisee par l'algorithme de Choleski, 
C i.e. $A=LL^T$, avec $L$ matrice triangulaire inferieure
C a diagonale positive.
C La matrice $L$ est stockee sous forme bande dans le tableau
C a de taille nMat=(n-1)*nBand+n, ou nBand est la largeur
C de bande (supposee calculee au prealable) des matrices
C $A$ et $L$.
C
C Le calcul se fait en 2 etapes successives:
C - une etape de descente, ou on resoud le systeme $Ly=g$,
C - une etape de remontee, ou on resoud le systeme $L^Tx=y$.
C Il n'est pas necessaire d'uiliser 2 tableaux 
C supplementaires pour stocker y et x; le tableau g(n) 
C represente au debut le second membre du systeme, puis  
C successivement, au cours du calcul, les solutions y
C puis x.
C
C
       DIMENSION a(*),g(n)
C
       DO 1 i=1,n
         som=g(i)
         i1=(i-1)*nBand
         jMin=max(1,i-nBand)
         DO 2 j=jMin,i-1
           som=som-a(i1+j)*g(j)
2        CONTINUE
         g(i)=som/a(i1+i)
1      CONTINUE
C
       DO 11 i1=1,n
         i=n-i1+1
         som=g(i)
         jMax=min(n,i+nBand)
         DO 12 j=i+1,jMax
           j1=(j-1)*nBand
           som=som-a(j1+i)*g(j)
12         CONTINUE
         g(i)=som/a((i-1)*nBand+i)
11     CONTINUE
       END
