C Fichier: BcalMat.for
C 
C---------------------------------------------------
      SUBROUTINE BcalMat(q,me,aire,nt,ns,a,nBand)
C---------------------------------------------------
C Calcule et stocke sous forme bande dans le tableau
C monodimensionnel a la partie triangulaire inferieure
C de la matrice A du systeme lineaire associee a la
C resolution du probleme modele (Laplacien avec 
C conditions aux limites de Dirichlet) ecrit sous 
C forme variationnelle. La largeur de bande nBand 
C est supposee etre connue.
C
C Le calcul des elements Aij de la matrice est fait  
C pour tous les indices de sommets i et j, meme si  
C ce sont des sommets frontieres. Il faut donc, apres 
C ce sous-programme, faire appel au sous-programme  
C BDirCL de maniere a recuperer les bonnes conditions 
C aux limites de Dirichlet sur la frontiere du domaine.
C
C Le calcul effectif des coefficients Aij se fait 
C triangle par triangle, en ne calculant, sur chaque 
C triangle Tk, que les contributions elementaires associees
C a deux indices locaux consecutifs il et jl de sommets 
C de ce triangle (on a alors: i=me(il,k), j=me(jl,k)) 
C ainsi que les elements diagonaux (jl=il). Ceci resulte
C de la symetrie de la matrice A. 
C Pour i different de j, le resultat de ce calcul est
C stocke dans le tableau a a un indice ij=(i-1)*nBand+j
C ou ji=(j-1)*nBand+i, selon les cas:i>j ou j>i.
C
C
C     entree:
C     -------
C       triangulation: ns,nt,q(2,ns),me(3,nt) et 
C         aire(nt);
C       nBand: largeur de bande de la matrice, calculee
C         precedemment par le sous-programme calBand
C         du fichier utilP1.for.
C     
C     sortie:
C     -------
C       a(nMat), nMat=(ns-1)*nBand+ns: tableau contenant 
C         les elements de la matrice A; plus precisement, 
C         l'element Aij est stocke dans le tableau a 
C         a l'indice ij=(i-1)*nBand+j.
C
C 
C 
      DIMENSION q(2,ns), me(3,nt),aire(nt),a(*)
C
      nMat=nBand*(ns-1)+ns
      DO 1 i=1,nMat
  1     a(i)=0.0
C
      DO 2 k=1,nt
        DO 2 il=1,3
         i=me(il,k)
         jl=mod(il,3)+1
         j=me(jl,k)
         xmatElem=elemMat(k,il,jl,aire(k),q,me)
         IF (i.GT.j)THEN
           ij=(i-1)*nBand+j
           a(ij)=a(ij)+xmatElem
         ELSE
           ji=(j-1)*nBand+i
           a(ji)=a(ji)+xmatElem
         ENDIF
         ii=(i-1)*nBand+i
         xmatElem= elemMat(k,il,il,aire(k),q,me)
         a(ii)=a(ii)+ xmatElem
2     CONTINUE
      END
C
C
C----------------------------------------------------------
      SUBROUTINE BDirCL(a,phi0,ns,ng,ff,cte,nBand)
C----------------------------------------------------------
C Ce sous-programme permet de recuperer les bonnes 
C conditions aux limites de Dirichlet pour les
C sommets frontieres.
C
C Pour chaque indice i de sommet frontiere ($ng(i) \ne 0$),
C on modifie simultanement le i eme element diagonal
C de la matrice A, en le remplacant par une tres grosse
C constante cte, et la i eme composante du second membre
C ff en la remplacant par le produit cte*phi0, ou phi0
C est la condition a la limite de Dirichlet. De cette 
C facon, la i eme ligne du systeme lineaire s'ecrit, apres
C simplification par la constante cte,
C   phi(i) + termes petits = phi0(i), 
C et la condition a la limite est satisfaite a un 
C terme d'ordre (1/cte) pres.
C
C
C     entree:
C     -------
C       triangulation: ns,ng(ns);
C       matrice a(*) et sa largeur de bande nBand;
C       condition a la limite de Dirichlet stockee dans le
C         tableau phi0 (N.B. ce tableau est suppose etre 
C         rempli pour tous les sommets);
C       second membre ff(ns) du systeme lineaire;
C       tres grosse constante cte.
C     
C     sortie:
C     -------
C       matrice a et second membre ff modifies.
C
C
      DIMENSION a(*),phi0(ns),ng(ns),ff(ns)
C
      DO 1 i=1,ns
      IF(ng(i).NE.0)THEN
        a((i-1)*nBand+i)=cte
        ff(i)=cte*phi0(i)
      ENDIF
1     CONTINUE
      END

