C Fichier MCholes.for
C
C-------------------------------------------
      SUBROUTINE MSolvSys(n,z,al,il,jl)
C-------------------------------------------
C Ce sous-programme calcule la solution x du systeme 
C $Ax=z$, ou A est une matrice symetrique definie positive 
C de taille n*n factorisee par l'algorithme de Choleski, 
C i.e. $A=LL^T$, avec $L$ matrice triangulaire inferieure
C a diagonale positive.
C La matrice $L$ est stockee sous forme Morse dans le tableau
C al avec les tableaux auxiliaires il et jl.
C
C Le calcul se fait en 2 etapes successives:
C - une etape de descente, ou on resoud le systeme $Ly=z$,
C - une etape de remontee, ou on resoud le systeme $L^Tx=y$.
C Il n'est pas necessaire d'uiliser 2 tableaux 
C supplementaires pour stocker y et x; le tableau z(n) 
C represente au debut le second membre du systeme, puis  
C successivement, au cours du calcul, les solutions y
C puis x.
C
C
      DIMENSION z(*),al(*),jl(*),il(*)
C
      DO 1 i=1,n
        som=z(i)
        DO 2 k=il(i), il(i+1)-2
 	        som=som-al(k)*z(jl(k))
2       CONTINUE
 	      z(i)=som/al(il(i+1)-1)
1     CONTINUE
C
      DO 3 i=1,n
3       z(i) = z(i)/al(il(i+1)-1)
      DO 4 i1=1,n
        i=n+1-i1
        DO 4 k=il(i), il(i+1)-2
          jlk=jl(k)
          aljj=al(il(jlk+1)-1)
          z(jlk) = z(jlk) - al(k)*z(i)/aljj
4       CONTINUE
      END
C
C
C----------------------------------------------------
      SUBROUTINE MCholeski(n,aa,ia,ja,al,il,jl,nBand)
C----------------------------------------------------
C Ce sous-programme calcule la factorisee de Choleski L
C d'une matrice symetrique definie positive A de taille
C n*n, dont seule la partie triangulaire inferieure est 
C stockee sous forme Morse dans le tableau aa, avec les
C tableaux auxiliaires ia et ja. 
C La factorisee de Choleski est stockee sous forme 
C Morse dans le tableau al, avec les tableaux 
C auxiliaires il et jl, qu"on construit en premier.
C La factorisation se fait ensuite colonne par 
C colonne, en commencant par le calcul de l"element
C diagonal.
C On suppose connue la largeur de bande nBand.
C
C N.B. Cette factorisation suppose que les elements
C du tableau aa verifient:
C $jl(k+1)>jl(k),\forall k\in[il(i),il(i+1)-2], \quad \forall i$
C
C
C
C     entree:
C     -------
C     largeur de bande nBand de la matrice A(n,n),i.e.
C       $nBand = max\{|i-j|: i>j, A(i,j) \ne 0\}$;
C     tableaux aa,ia et ja de stockage Morse de 
C       la matrice A
C
C
C     sortie:
C     -------
C      tableaux al,il et jl de stockage Morse de la
C        factorisee L de Choleski de A.
C
C
      DIMENSION aa(*),ia(*),ja(*)
      DIMENSION al(*),il(*),jl(*)
      k=0
      DO 41 i=1,n
        il(i)=k+1
        if(ja(ia(i)).GT.i)write(*,*)i,ia(i),ja(ia(i)),k
        DO 41 j=ja(ia(i)),i
          k=k+1
          jl(k)=j
41    CONTINUE
      il(n+1)=k+1
C
      DO 34 j=1,n
        aljj=aa(ia(j+1)-1)
        DO 33 k=il(j), il(j+1)-2
          aljj=aljj-al(k)**2
33      CONTINUE
        aljj=sqrt(aljj)		
        al(il(j+1)-1)=aljj
        iMax=min(j+nBand,n)
        DO 34 i=j+1,iMax
          DO 32 k=ia(i),ia(i+1)-2
            IF(ja(k).GE.j)GOTO 37
32        CONTINUE
37        IF(ja(k).NE.j) THEN 
            alij=0.
          ELSE
            alij=aa(k)
          ENDIF
          lk=0
          ik=il(i)
          jk=il(j)
          maxik=il(i+1)-2
          maxjk=il(j+1)-2
35        IF(ik.LE.maxik) THEN
            jlik=jl(ik)
            IF(jlik.EQ.j) lk=ik
36          IF(jk.LE.maxjk) THEN
              jljk=jl(jk)
              IF(jlik.GT.jljk)THEN
                jk=jk+1
                GOTO 36
              ELSEIF(jlik.LT.jljk)THEN
                ik=ik+1
                GOTO 35
              ELSE
                alij=alij-al(ik)*al(jk)
                ik=ik+1
                jk=jk+1
                GOTO 35
              ENDIF
            ENDIF
          ENDIF
          IF(lk.NE.0)al(lk)=alij/aljj
34    CONTINUE
      END
