C Fichier: McalMat.for
C
C---------------------------------------------------
      SUBROUTINE McalMat(q,me,ng,aire,nt,ns,aa,ia,ja)
C---------------------------------------------------
C Ce sous-programme calcule et stocke les elements
C de la partie triangulaire inferieure de la matrice
C A du systeme lineaire associe a la resolution du 
C Laplacien avec conditions aux limites de Dirichlet 
C ecrit sous forme variationnelle et discretise par
C elements finis $P^1$. Il calcule egalement la 
C taille effective nElemA du tableau aa ou se trouve
C stockee la matrice A sous forme Morse.
C
C L'element Aij de cette matrice est stocke dans le 
C tableau monodimensionnel a a un indice k; ja(k)
C contient l"indice de colonne j de cet element. 
C Le tableau ia est defini par: ia(i) est l"indice
C de stockage dans le tableau a du premier element
C de la i eme ligne de A effectivement stocke dans 
C ce tableau. 
C En convenant de poser ia(ns+1)=nElemA+1, 
C l"element Aii de la matrice A est stocke dans 
C le tableau aa a l"indice ia(i+1)-1.
C Au debut du calcul, le tableau ia a une autre
C signification: ia(i) est le nombre de sommets
C strictement voisins du point d"indice i
C (i.e. exception faite du point lui-meme).
C
C
C     entree:
C     ------
C     triangulation (ns,nt,q,me,ng) et les aires des
C       triangles stockees dans le tableau aire.
C
C     sortie:
C     ------
C     tableaux aa,ia,ja; taille nElemA du tableau aa
C
C    
C     remarques importantes:
C     ---------------------
C     - on suppose les triangles orientes positivement
C     - le tableau ng sert a detecter les aretes 
C       frontieres de la maniere suivante: une arete
C       est frontiere si et seulement si ses 2 
C       extremites sont des sommets frontieres.
C
C     Il faut donc verifier que la triangulation 
C     etudiee satisfait ces deux conditions.
C
C
      DIMENSION q(2,ns),me(3,nt),ng(ns),aire(nt)
      DIMENSION aa(*),ia(*),ja(*)
C
      DO 110 i=1,ns
110     ia(i)=0
      DO 111 k = 1,nt
        DO 111 il = 1,3
          ilp = mod(il,3)+1
          i=me(il,k)
          ip=me(ilp,k)
          IF (ip.LE.i) THEN
             ia(i) = ia(i)+1
          ELSEIF ((ng(i).NE.0).AND.(ng(ip).NE.0)) THEN 
             ia(ip) = ia(ip)+1
          ENDIF	
111     CONTINUE
C
C ia(i) est le nombre de sommets voisins stricts 
C du sommet d"indice i.
C
      nElemA=0
      DO 211 i = 1,ns
        ivois=ia(i)
        ia(i)= nElemA + 1
        nElemA = nElemA + ivois + 1
211   CONTINUE
      ia(ns+1)=nElemA+1
C
C ia est construit; reste a construire aa et ja.
C
      DO 1 k = 1, nElemA
      aa(k)=0.
1     ja(k)=0
      DO 2 kt = 1,nt
        DO 2 il=1,3
          i = me(il,kt)
          DO 2 jl=1,3
            j=me(jl,kt)
            k=0
            IF (i.EQ.j) THEN
              k=ia(i+1)-1
              ja(k)=i
            ELSEIF (i.GT.j) THEN
              DO 3 l= ia(i), ia(i+1)-2
C
C j ne peut pas etre ailleurs
C
                IF (ja(l).EQ.0) GOTO 4
                IF ( ja(l) .EQ. j )THEN
                  k=l
                  GOTO 4
                ENDIF
3             CONTINUE
	     write(*,*)'erreur dans McalMat.for au point i,j=',i,j
4             IF (k.EQ.0) THEN 
                k = l
                ja(l) = j 
              ENDIF
            ENDIF
            IF(k.NE.0)aa(k)=aa(k)+
     >        elemMat(kt,il,jl,aire(kt),q,me)
2     CONTINUE
      END
C
C
C----------------------------------------------------------
      SUBROUTINE MDirCL(aa,ia,phi0,ns,ng,ff,cte)
C----------------------------------------------------------
C Ce sous-programme permet de recuperer les bonnes 
C conditions aux limites de Dirichlet pour les
C sommets frontieres.
C
C Pour chaque indice i de sommet frontiere ($ng(i)\ne 0$),
C on modifie simultanement le i eme element diagonal
C de la matrice A, en le remplacant par une tres grosse
C constante cte, et la i eme composante du second membre
C ff en la remplacant par le produit cte*phi0, ou phi0
C est la condition a la limite de Dirichlet. De cette 
C facon, la i eme ligne du systeme lineaire s'ecrit, apres
C simplification par la constante cte,
C   phi(i) + termes petits = phi0(i), 
C et la condition a la limite est satisfaite a un 
C terme d'ordre (1/cte) pres.
C
C
C     entree:
C     -------
C       triangulation: ns,ng(ns);
C       matrice aa(*) et tableau ia(*);
C       condition a la limite de Dirichlet stockee dans le
C         tableau phi0 (N.B. ce tableau est suppose etre 
C         rempli pour tous les sommets);
C       second membre ff(ns) du systeme lineaire;
C       tres grosse constante cte.
C     
C     sortie:
C     -------
C       matrice aa et second membre ff modifies.
C
C
      DIMENSION aa(*),ia(ns+1),phi0(ns),ng(ns),ff(ns)
      DO 1 i=1,ns
        IF(ng(i).NE.0)THEN
          aa(ia(i+1)-1)=cte
          ff(i)=cte*phi0(i)
        ENDIF
1     CONTINUE
      END
C
C
