C Fichier GradCP.for
C
C----------------------------------------------------------
      SUBROUTINE GradC(n,aa,ia,ja,al,b,x,g,h,cg,kMax,eps)
C----------------------------------------------------------
C Ce sous-programme resout le systeme lineaire $Ax=b$ par
C gradient conjugue preconditionne, avec au plus kMax 
C iterations. Le test d"arret est: $|Ax^k-b|<eps|Ax^0-b|$.
C La matrice $A$ du systeme lineaire, de taille $n \times n$
C est stockee sous forme Morse dans le tableau aa;
C ce stockage necessite aussi les tableaux auxiliaires
C d"indices ia et ja, comme explique dans le chapitre III.
C La matrice $C$ de preconditionnement est la factorisation
C de Choleski incomplete de $A$ ($C=L^* (L^*)^{T}$)  et la
C matrice $L^*$ est stockee sous forme Morse dans le 
C tableau al, avec les memes tableaux d"indices ia et 
C ja que $A$.
C
C     entree:
C     --------
C       matrice $A$ du systeme stockee sous forme Morse 
C         dans les tableaux (aa,ia,ja),
C       factorisee de Choleski incomplete de $A$ stockee
C         dans les tableaux (al,ia,ja);
C       second membre de l"equation stocke dans le 
C         tableau b;
C       nombre maximal kMax d"iterations, precision eps
C         du calcul;
C
C     N.B.: le calcul du second membre ne servant qu"a 
C       l"etape d"initialisation, le tableau b est ensuite
C       ecrase et utilise comme tableau auxiliaire de  
C       calcul, en particulier pour le stockage du produit 
C       de la matrice $A$ par $h$, puis pour le stockage 
C       de la solution y du systeme $C y= A h$.
C       
C                            
C     sortie:
C     ------- 
C       x:tableau contenant les valeurs de la solution 
C         en chacun des sommets du maillage;
C       gcg:scalaire contenant le carre de la norme de
C         $g$ (en pratique ce nombre doit decroitre au 
C         cours des iterations).
C                            
C     autres tableaux:
C     ----------------
C       g, cg: tableaux contenant les valeurs de $g^n$ et
C         de $C g^n$ respectivement;
C       h: tableau contenant la direction de descente $h^n$
C         (au tout debut du programme, ce tableau sert a 
C          stocker le produit de $A$ par $x^0$);
C       
C
C     sous-programmes:
C     ---------------- 
C       IAmul:effectue le produit d"une matrice par 
C         un vecteur, la matrice etant stockee sous
C         forme Morse,
C       ISolvSys:resolution du systeme lineaire 
C         $L L^T x=y$ par descente remontee, pour
C         une matrice $L$ stockee sous forme Morse;
C       Ces deux sous-programmes sont dans le fichier
C       GCPLaplac.for.
C
C
      DIMENSION x(*),b(*),g(*),h(*),cg(*)
      DIMENSION aa(*),ja(*),ia(*),al(*)
C
      DO 1 k=1, kMax
C
        IF (k.EQ.1) THEN
          CALL IAmul(n,aa,ia,ja,x,h)
          DO 2 i=1,n
            cg(i)=b(i)-h(i)
            g(i)=cg(i)
2         CONTINUE
          CALL ISolvSys(n,g,al,ia,ja)
          gcg=0.
          DO 3 i=1,n
            gcg=gcg + g(i)*cg(i)
            h(i)=g(i)
3         CONTINUE
        ENDIF
C
        CALL IAmul(n,aa,ia,ja,h,b)
        hAh=0.
        DO 10 i=1,n		
          hAh=hAh+h(i)*b(i)
10      CONTINUE
        ro=gcg/hAh
        DO 11 i=1,n
          cg(i) = cg(i) -ro*b(i)
11      CONTINUE
        CALL ISolvSys(n,b,al,ia,ja)
        DO 12 i=1,n
          x(i)=x(i)+ro*h(i)
          g(i)=g(i)- ro*b(i)
12      CONTINUE
        gamma=gcg
        gcg=0
        DO 13 i=1,n
          gcg=gcg + g(i)*cg(i)
13      CONTINUE
C
        write(*,*) 'iteration:',k,'gcg=',gcg
        IF(k.EQ.1) eps=eps*gcg
        IF (gcg.LT.eps) RETURN
        gamma=gcg/gamma
        DO 14 i=1,n
          h(i)=g(i)+gamma*h(i)
14      CONTINUE
C
1     CONTINUE
      END

