C Fichier: gelem.for
C
C-----------------------------------------------------
      SUBROUTINE gSMbre(me,f,ns,nt,aire,ff,ng,q)
C-----------------------------------------------------
C Ce sous-programme calcule le second membre 
C     $ff(i)=\int_\Omega f w^i + \int_{\Gamma_2} g w^i$
C du systeme lineaire a resoudre, et stocke le resultat
C dans le tableau ff.
C
C Ce calcul se fait en plusieurs etapes:
C
C - on calcule d"abord la deuxieme integrale, car
C   en entree de ce sous-programme le tableau f
C   contient les valeurs de $g$ en chacun des sommets
C   du maillage (ce tableau f pourra etre ensuite ecrase);
C   
C - on lit dans un fichier, dont le nom est donne par 
C   l"utilisateur au moment de l"execution du programme,
C   les valeurs du second membre $f$, qu'on stocke dans le 
C   tableau f; cette lecture se fait par l"intermediaire
C   du sous-programme " litF "du fichier utilP1.for;
C
C - on calcule ensuite la premiere integrale.
C
C
C     entree:
C     -------
C       triangulation (ns,nt,q,me,ng,aire),
C       tableau f contenant les valeurs de g en chacun des
C         sommets du maillage.
C       
C     sortie:
C     -------
C       le tableau f contient les valeurs du second membre
C         f de l"EDP en chacun des sommets du maillage,
C       le tableau ff contient le second membre du systeme 
C         lineaire a resoudre.
C         
C
      DIMENSION q(2,*),me(3,*),ng(*),aire(*)
      DIMENSION f(*),ff(*)
      DIMENSION isuiv(4)
      CHARACTER*20 nomfic
C
      DATA isuiv/2,3,1,2/
      DO 2 j=1,ns
        ff(j)=0.
2     CONTINUE
C
      DO 3 k = 1,nt
        DO 3 il=1,3
          i=me(il,k)
          IF (ng(i).NE.0) THEN
            ip=me(isuiv(il),k)
            ipp=me(isuiv(il+1),k)
            IF (ng(ip).NE.0) THEN
              dx=q(1,i)-q(1,ip)
              dy=q(2,i)-q(2,ip)
              dist=sqrt(dx*dx+dy*dy)
              ff(i)=ff(i)+(2*f(i)+f(ip))*dist/6.
            ENDIF
            IF (ng(ipp).NE.0) THEN
              dx=q(1,i)-q(1,ipp)
              dy=q(2,i)-q(2,ipp)
              dist=sqrt(dx*dx+dy*dy)
              ff(i)=ff(i)+(2*f(i)+f(ipp))*dist/6.
            ENDIF
          ENDIF  
3     CONTINUE
C
C
C
      write(*,*) 'nom du fichier de stockage de f'
      write(*,*) '(second membre de l"EDP)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
C
      DO 4 k = 1,nt
        airek=aire(k)
        DO 4 il=1,3
          i=me(il,k)
          ip=me(isuiv(il),k)
          ipp=me(isuiv(il+1),k)
          ff(i)=ff(i)+(2*f(i)+f(ip)+f(ipp))*airek/12.
4     CONTINUE
      END
C
C
C---------------------------------------------------------
      FUNCTION geMat(k,il,jl,airek,q,me,ng,ns,nt,AM,iM,a0,
     >ia0,a1,ia1)
C---------------------------------------------------------
C Cette fonction calcule la contribution elementaire sur le
C triangle d"indice k du coefficient $A_{ij}$ de la matrice
C du systeme lineaire a resoudre, dans le cas d'un
C operateur elliptique symetrique du second ordre general
C       $a_{0}\varphi  -\nabla .(M\nabla \varphi)$,
C avec des conditions aux limites de type Robin 
C sur la frontiere $\Gamma_2 $
C       $a_{1}\varphi + n^{T}M\nabla \varphi =g. $ 
C
C Le calcul se fait ensuite en 3 etapes:
C - calcul de $\int_\Omega (\nabla w)^{T}M\nabla \varphi $,
C - calcul de $\int_\Omega a_{0}\varphi w $,
C - calcul de $\int _{\Gamma} a_{1}\varphi w$.
C
C Chacun de ces calculs est fait en tenant compte de la 
C structure particuliere eventuelle du probleme.
C Nous rappelons que:
C - $iM=0$ si $M=Id$,
C - $iM \le 1$ si $M=M_{11} Id$,
C - $iM \le 2$ si $M$ est diagonale,
C - $iM=3$ dans les autres cas.
C Nous avons aussi: $ia0=0$ si $a_0$ est la fonction nulle,
C et $ia0=1$ sinon; il en est de meme pour $ia1$ en fonction
C de $a_1$.
C
C Le tableau AM contient ici les valeurs moyennes $m$ sur 
C chaque triangle de la matrice $M$ (i.e. le tiers de la 
C somme des valeurs prises par cette matrice en chacun des
C sommets du triangle). Il est rempli de la maniere 
C suivante:
C - les nt premiers indices correspondent au stockage
C   des coefficients $m_{11}$,
C - les nt suivants a $m_{22}$,
C - enfin les nt derniers aux coefficients extra-diagonaux
C   $m_{12}$.
C
C
C N.B. Ce sous-programme suppose les triangles orientes 
C      positivement. 
C
C
      DIMENSION q(2,*),me(3,*),ng(*),isuiv(4)
      DIMENSION AM(*),a0(*),a1(*)
C
      DATA isuiv/2,3,1,2/
C
C etape 1: matrice $M$ de l"EDP
C
      nkt=k-nt
      IF (iM.GT.0) THEN
        nkt=nkt+nt
        AM11=AM(nkt)
      ELSE
        AM11=1.
      ENDIF
      IF (iM.GT.1) THEN
        nkt=nkt+nt
        AM22=AM(nkt)
      ELSE
        AM22=AM11
      ENDIF
C
      i=me(il,k)
      ip=me(isuiv(il),k)
      ipp=me(isuiv(il+1),k)
      j=me(jl,k)
      jp=me(isuiv(jl),k)
      jpp=me(isuiv(jl+1),k)
      gradWi1= q(2,ip)-q(2,ipp)
      gradWi2= q(1,ipp)-q(1,ip)
      gradWj1= q(2,jp)-q(2,jpp)
      gradWj2= q(1,jpp)-q(1,jp)
      elem=AM11*gradWi1*gradWj1+AM22*gradWi2*gradWj2
C
      IF (iM.EQ.3) THEN
        nkt=nkt+nt
        elem=elem+AM(nkt)*(gradWi1*gradWj2+gradWi2*gradWj1)
      ENDIF
      elem=elem/(4*airek)
C
C etape 2: terme constant $a_0$ de l"EDP
C
      IF (ia0.EQ.1) THEN
        a0k=(a0(i)+a0(ip)+a0(ipp))/3.
        IF (i.EQ.j) THEN 
          elem=elem+a0k*airek/6.
        ELSE
          elem=elem+a0k*airek/12.
        ENDIF
      ENDIF
C
C etape 3: terme $a_1$ de la condition a la limite
C
      IF (ia1.EQ.1) THEN
        IF (ng(i).NE.0) THEN
          IF (i.EQ.j) THEN
            IF (ng(ip).NE.0) THEN
              dx=q(1,i)-q(1,ip)
              dy=q(2,i)-q(2,ip)
              dist=sqrt(dx*dx+dy*dy)
              elem=elem+(3*a1(i)+a1(ip))*dist/12.
            ENDIF
            IF (ng(ipp).NE.0) THEN
              dx=q(1,i)-q(1,ipp)
              dy=q(2,i)-q(2,ipp)
              dist=sqrt(dx*dx+dy*dy)
              elem=elem+(3*a1(i)+a1(ipp))*dist/12.
            ENDIF
          ELSEIF (ng(j).NE.0) THEN
            dx=q(1,i)-q(1,j)
            dy=q(2,i)-q(2,j)
            dist=sqrt(dx*dx+dy*dy)
            elem=elem+(a1(i)+a1(j))*dist/12.
          ENDIF
        ENDIF
      ENDIF
      geMat=elem
      END
C
C
C-----------------------------------------------------
      SUBROUTINE gDirCL(aa,ia,ff,phi0,ng,ns,cte)
C-----------------------------------------------------
C Ce sous-programme permet d"assurer la condition a la
C limite de Dirichlet sur la partie $\Gamma_1$ du bord
C du domaine correspondant aux sommets $q^i$ ne 
C verifiant pas phi0(i) = 0.
C 
C Pour chacun de ces points, on modifie simultanement:
C - l"element diagonal de aa en le remplacant par une
C     tres grosse constante cte,
C - ff(i) en le remplacant par le produit phi0(i)*cte.
C
C
      DIMENSION aa(*),ia(*),phi0(*),ng(*),ff(*)
C
      DO 1 i = 1,ns
        IF ((ng(i).NE.0).AND.(phi0(i).NE.0)) THEN 
          aa(ia(i+1)-1)=cte
          ff(i)=phi0(i)*cte
        ENDIF
1     CONTINUE
      END
C
C
