C  fichier LplcMinE.for
C
       PROGRAM Laplace
C      ---------------  
C
C Ce programme resout le probleme modele,i.e. le Laplacien
C avec conditions aux limites de Dirichlet:
C $-\Delta \phi=f$, $\phi$ donne sur le bord.
C 
C La methode employee est celle de la minimisation de 
C l'energie par gradient conjugue sur tous les 
C sommets du maillage (Algorithme 2 modifie). 
C
C On definit un nombre maximal mMax d'iterations et 
C une precision eps;l'algorithme s'arrete quand le 
C nombre d'iterations effectuees est plus grand que 
C mMax ou quand la precision souhaitee est atteinte,
C i.e. $ ||grad E^m||^2< $ eps $ ||grad E^0||^2 $, ou 
C $ ||grad E^m|| $ est le gradient de la fonctionnelle 
C energie a l'iteration m.
C
C     entree:
C     --------- 
C       triangulation (ns,nt,q,me,ng) et donnees du 
C         probleme (second membre f de l'equation de 
C         Laplace et condition a la limite de Dirichlet 
C         stockee dans le tableau phi) lues dans le 
C         fichier nomfic par l'intermediaire des sous-
C         programmes litT et litF;
C       parameter:nsMax (nombre maximal de sommets) et 
C         ntMax (nombre maximal de triangles) de la 
C         triangulation. 
C       precision eps et nombre maximal d'iterations 
C         mMax,
C       initialisation du tableau hConjm et de g2hm.
C                            
C     sortie:
C     ------- 
C       phi:tableau contenant les valeurs de la solution 
C         approchee en chacun des sommets du maillage,  
C       g2h:carre de la norme du gradient
C         de la fonctionnelle. 
C                            
C     tableaux:
C     --------- 
C     maillage:
C       q(2,nsMax):coordonnees des sommets du maillage,
C       me(3,ntMax):numerotation des sommets des
C         triangles,
C       ng(nsMax):tableau indiquant si un sommet est 
C         interne (ng(i)=0) ou non.
C
C     donnees:
C       f(nsMax):second membre du Laplacien,
C       phi(nsMax):au depart, conditions aux limites de 
C         Dirichlet (ensuite, valeurs de la solution 
C         approchee).
C
C     autres tableaux:
C       gradE(nsMax):gradient, en chacun des sommets du 
C         maillage, de la fonctionnelle E,
C       hConj(nsMax):nouvelle direction de descente,
C       hConjm(nsMax):ancienne direction de descente.
C
C     scalaires:
C     ----------
C       g2h,g2hm:carre de la norme du gradient de la 
C         fonctionnelle energie a l'iteration en cours 
C         (g2h) et a l'iteration precedente (g2hm),
C       gamma:coefficient $\gamma={||grad E^m||^2\over ||grad E^{m-1}||^2}$, 
C       ro:reel qui minimise l'energie sur la droite  
C         passant par phi et de vecteur directeur hConj.
C
C     fichiers:
C     ---------
C       nomfic:nom du fichier de stockage du maillage, 
C         des donnees du probleme modele (terme source du 
C         Laplacien et condition a la limite de Dirichlet),
C         ou de la solution approchee, en vue, par exemple,
C         d'une exploitation graphique.
C
C     sous-programmes:
C     ---------------- 
C       litT:lecture du maillage,
C       litF:lecture des donnees du probleme,
C       dEdPhi:calcul du gradient gradE de E et du carre  
C         g2h de la norme euclidienne du gradient de E,
C       ecritF:ecriture de la solution approchee.
C
C     fonctions:
C     ----------
C       rho:calcul de $\rho$,
C       E:calcul de l'energie,
C       deter:calcul d'un determinant particulier.
C        
C     bibliotheque (sous-programmes et fonctions):
C     ----------------------------------------
C       minEe.for:fichier contenant la fonction E,
C       minEdE.for:fichier contenant le sous-programme
C         dEdPhi,
C       minErho.for:fichier contenant la fonction rho,
C       utilP1.for:fichier contenant la fonction deter,
C         les sous-programmes litT,litF et ecritF. 
C                  
C
      PARAMETER (nsMax=2000, ntMax=4000)            
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)		
      DIMENSION phi(nsMax),f(nsMax)				
      DIMENSION gradE(nsMax),hConj(nsMax),hConjm(nsMax)			
      CHARACTER*20 nomfic			
C
      DATA eps/1.0e-14/			
      DATA mMax/300/
C				
      DATA hConjm/nsMax*0./ 
      g2hm=1.0				
C
      write(*,*) 'nom du fichier de stockage du maillage'
      read(*,'(a)') nomfic
      CALL litT(nomfic,ns,nt,q,ng,me,ngt)
      write(*,*) 'nom du fichier de stockage de f'
      write(*,*) '(second membre du Laplacien)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
      write(*,*) 'nom du fichier de stockage de phi'
      write(*,*) '(conditions aux limites de Dirichlet)'
      read(*,'(a)') nomfic  
      CALL litF(nomfic,ns,phi)

      write(*,*)'iter  energie        ro          g2h'
C
      DO 1 m = 1,mMax	
        CALL dEdPhi(q,me,ng,phi,f,ns,nt,gradE,g2h)
        IF(m.EQ.1) eps = eps*g2h
        gamma=g2h/g2hm	
        DO 2 i=1,ns
          hConj(i)=-gradE(i)+gamma*hConjm(i)
2       CONTINUE	
        ro = rho(q,me,ns,nt,phi,gradE,hConj,hConjm)
        E1 = E(q,me,phi,f,ns,nt)
        write(*,*) m,E1,ro,g2h
        IF(g2h.LT.eps) GOTO 4
        DO 3 i = 1,ns
          phi(i) = phi(i)+ ro* hConj(i)
          hConjm(i)=hConj(i)
3       CONTINUE
        g2hm=g2h
1     CONTINUE
C
4     CONTINUE
C	
      write(*,*) 'nom du fichier de stockage de la '
      write(*,*) 'solution approchee'
      read(*,'(a)') nomfic
      CALL ecritF(nomfic,ns,phi)  
      PAUSE
      END
C
C
C********************************* 
C 
C     BIBLIOTHEQUE
C    
C********************************* 
C
C                            
      INCLUDE 'minEe.for'
      INCLUDE 'minEdE.for'
      INCLUDE 'minErho.for'
      INCLUDE 'utilP1.for'
