C Fichier: BLaplace.for 
C
C 
      PROGRAM BLaplace
C     -----------------
C Ce programme resout le probleme modele,i.e. le Laplacien
C avec conditions aux limites de Dirichlet :
C $-\Delta \Phi=f, \Phi$ donne sur le bord.
C
C Le probleme est ecrit sous forme variationnelle, puis 
C discretise par elements finis $P^1$. La solution appro-
C chee en chacun des sommets de la triangulation est  
C determinee en resolvant un systeme lineaire, systeme  
C qu'on resout par une methode directe, en factorisant 
C la matrice sous forme de Choleski.
C La matrice est stockee sous forme bande.
C La matrice A, puis sa factorisee de Choleski L,
C sont stockees dans le meme tableau a.
C
C
C     entree:
C     --------- 
C       triangulation (ns,nt,q,me,ng) et donnees du
C         probleme (second membre f de l'equation
C         de Laplace et condition a la limite de 
C         Dirichlet stockee dans le tableau phi0)
C         lues dans le fichier nomfic par l'intermediaire
C         des sous-programmes litT et litF;
C       parameter:nsMax (nombre maximal de sommets), 
C         ntMax (nombre maximal de triangles) de la
C         triangulation, nBandMax (largeur de bande
C         maximale) et nMatMax=(nsMax-1)*nBandMax
C         +nsMax (taille maximale du tableau a).
C                            
C     sortie:
C     ------- 
C       ff:tableau contenant les valeurs de la solution 
C         approchee en chacun des sommets du maillage,
C         stocke dans un fichier (dont le nom est precise
C         par l"utilisateur) en vue d"une eventuelle
C         exploitation graphique.
C                            
C     tableaux:
C     --------- 
C     maillage:
C       q(2,nsMax):coordonnees des sommets du maillage,
C       me(3,ntMax):numerotation des sommets des 
C         triangles,
C       ng(nsMax):tableau indiquant si un sommet est sur 
C          la frontiere ($ng(i) \ne 0$) ou non ($ng(i)=0$).
C
C     donnees:
C       f(nsMax):second membre du Laplacien,
C       phi0(nsMax):conditions aux limites de Dirichlet
C
C     autres tableaux:
C       a(nMatMax):tableau de stockage, sous forme bande, 
C         de la matrice A, en debut de programme, puis de
C         sa factorisee de Choleski L,
C       aire(ntMax):aire de chacun des triangles,
C       ff(nsMax):au depart, second membre du systeme
C         lineaire a resoudre, puis solution de ce 
C         systeme.
C
C     scalaires:
C     ----------
C       nBand: largeur de bande de la matrice,
C       nMat: dimension du tableau a,
C       cte: valeur de la tres grosse constante permettant
C         de satisfaire la condition a la limite de 
C         Dirichlet,
C       err: erreur en norme $H^1_0$ entre la solution 
C         exacte phi0 et la solution approchee ff.
C
C     fichiers:
C     ---------
C       nomfic:nom du fichier de stockage du maillage, 
C         des donnees du probleme modele (terme source du 
C         Laplacien et condition a la limite de Dirichlet),
C         ou de la solution approchee, en vue, par exemple,
C         d'une exploitation graphique.
C
C     sous-programmes:
C     ---------------- 
C       litT:lecture du maillage,
C       litF:lecture des donnees du probleme;
C       calBand:calcul de la largeur de bande nBand de
C         la matrice A;
C       calAire:calcul de l"aire de chaque triangle;
C       BcalMat:calcul et stockage, sous forme bande dans
C         le tableau a, de la matrice A;
C       SndMbre:calcul, et stockage dans le tableau ff, 
C         du second membre du systeme;
C       BDirCL:programme assurant la satisfaction des 
C         conditions aux limites de Dirichlet;
C       BCholeski:factorisation de Choleski de la matrice
C         A stockee sous forme bande;
C       BSolvSys:resolution du systeme lineaire par 
C         descente-remontee; utilise la factorisee L de
C         Choleski de A stockee dans le tableau a sous
C         forme bande;
C       ecritF:ecriture de la solution approchee.
C       
C
C     fonctions:
C     ----------
C       errH10:calcul de l"erreur $H^1_0$ entre
C         la solution exacte phi0 et la solution 
C         approchee ff.
C        
C     bibliotheque (sous-programmes et fonctions):
C     -------------------------------------------
C       utilP1.for:fichier contenant les fonctions
C         deter (pour le calcul des aires), errH10
C         et les sous-programmes litT,litF, ecritF
C         et calBand;
C       SndMbre.for:fichier contenant les sous-
C         programmes calAire et SndMbre;
C       elem.for:fichier contenant la fonction
C         elemMat;
C       BcalMat.for:fichier contenant les sous-
C         programmes BcalMat et BDirCL;
C       BCholes.for:fichier contenant les sous-
C         programmes BCholeski et BSolvSys.
C                  
C
      PARAMETER (nsMax=1800, ntMax=3400,nBandMax=70)            
      PARAMETER (nMatMax=(nsMax-1)*nBandMax+nsMax)           
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
      DIMENSION aire(ntMax)		
      DIMENSION phi0(nsMax),f(nsMax),ff(nsMax)
      DIMENSION a(nMatMax)
      CHARACTER*20 nomfic			
C
      DATA cte/1.0e25/			
C				
      write(*,*) 'nom du fichier de stockage du maillage'
      read(*,'(a)') nomfic
      CALL litT(nomfic,ns,nt,q,ng,me,ngt)
      write(*,*) 'nom du fichier de stockage de f'
      write(*,*) '(second membre du Laplacien)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
      write(*,*) 'nom du fichier de stockage de phi0'
      write(*,*) '(conditions aux limites de Dirichlet)'
      read(*,'(a)') nomfic  
      CALL litF(nomfic,ns,phi0)
C
      CALL calBand(me,ng,nt,ns,nBand)
      CALL calAire(q,me,nt,ns,aire)
      CALL BcalMat(q,me,aire,nt,ns,a,nBand)
      CALL SndMbre(me,f,ns,nt,aire,ff)
      CALL BDirCL(a,phi0,ns,ng,ff,cte,nBand)
      CALL BCholeski(ns,a,nBand)
      CALL BSolvSys(ns,a,ff,nBand)
C
      err=errH10(q,me,ff,phi0,ns,nt)
      write(*,*) 'erreur=',err
      write(*,*) 'nom du fichier de stockage de la '
      write(*,*) 'solution approchee'
      read(*,'(a)') nomfic
      CALL ecritF(nomfic,ns,ff)  
      PAUSE
      END
C
C
C**********************************
C
C     BIBLIOTHEQUE
C
C**********************************
C
C
      INCLUDE 'elem.for'
      INCLUDE 'BcalMat.for'
      INCLUDE 'BCholes.for'
      INCLUDE 'utilP1.for'
      INCLUDE 'SndMbre.for'


