C
      PROGRAM MLaplace
C
C Ce programme resout le probleme modele,i.e. le Laplacien
C avec conditions aux limites de Dirichlet :
C $-\Delta \Phi=f, \Phi$ donne sur le bord.
C
C Le probleme est ecrit sous forme variationnelle, puis 
C discretise par elements finis $P^1$. La solution appro-
C chee en chacun des sommets de la triangulation est  
C determinee en resolvant un systeme lineaire, systeme  
C qu'on resout par une methode directe, en factorisant 
C la matrice sous forme de Choleski.
C La matrice est stockee sous forme Morse.
C La matrice A et sa factorisee de Choleski L
C ne sont pas stockees dans le meme tableau.
C
C
C     entree:
C     --------- 
C       triangulation (ns,nt,q,me,ng) et donnees du
C         probleme (second membre f de l'equation
C         de Laplace et condition a la limite de 
C         Dirichlet stockee dans le tableau phi0)
C         lues dans le fichier nomfic par l'intermediaire
C         des sous-programmes litT et litF;
C       parameter:nsMax (nombre maximal de sommets), 
C         ntMax (nombre maximal de triangles) de la
C         triangulation, nBandMax (largeur de bande
C         maximale), nMLMax=(nsMax-1)*nBandMax
C         +nsMax (taille maximale du tableau al) et
C         nMAMax=8*nsMax (taille maximale du tableau
C         aa).
C                            
C     sortie:
C     ------- 
C       ff:tableau contenant les valeurs de la solution 
C         approchee en chacun des sommets du maillage,
C         stocke dans un fichier (dont le nom est precise
C         par l"utilisateur) en vue d"une eventuelle
C         exploitation graphique.
C                            
C     tableaux:
C     --------- 
C     maillage:
C       q(2,nsMax):coordonnees des sommets du maillage,
C       me(3,ntMax):numerotation des sommets des
C         triangles,
C       ng(nsMax):tableau indiquant si un sommet est sur 
C          la frontiere ($ng(i) \ne 0$) ou non ($ng(i)=0$).
C
C     donnees:
C       f(nsMax): second membre du Laplacien,
C       phi0(nsMax): conditions aux limites de Dirichlet
C
C     autres tableaux:
C       aa(nMAMax): tableau de stockage, sous forme Morse, 
C         de la matrice A,
C       ia(nsMax+1): indice de stockage dans le tableau aa
C         du premier element de la i eme ligne de A 
C         effectivement stocke dans aa,
C       ja(nMAMax): indice j de colonne de l"element 
C         Aij stocke;
C       al(nMLMax), il(nsMax+1),jl(nMLMax): meme
C         signification que les 3 tableaux precedents,
C         mais pour la factorisee de Choleski L de A;
C       aire(ntMax):aire de chacun des triangles,
C       ff(nsMax):au depart, second membre du systeme
C         lineaire a resoudre, puis solution de ce 
C         systeme.
C
C     scalaires:
C     ----------
C       nBand: largeur de bande de la matrice,
C       cte: valeur de la tres grosse constante permettant
C         de satisfaire la condition a la limite de 
C         Dirichlet,
C       err: erreur en norme $H^1_0$ entre la solution 
C         exacte phi0 et la solution approchee ff.
C
C     fichiers:
C     ---------
C       nomfic:nom du fichier de stockage du maillage, 
C         des donnees du probleme modele (terme source du 
C         Laplacien et condition a la limite de Dirichlet),
C         ou de la solution approchee, en vue, par exemple,
C         d'une exploitation graphique.
C
C     sous-programmes:
C     ---------------- 
C       litT:lecture du maillage,
C       litF:lecture des donnees du probleme;
C       calBand:calcul de la largeur de bande nBand de
C         la matrice A;
C       calAire:calcul de l"aire de chaque triangle;
C       McalMat:calcul et stockage, sous forme Morse dans
C         le tableau aa, de la matrice A;
C       SndMbre:calcul, et stockage dans le tableau ff, 
C         du second membre du systeme;
C       MDirCL:programme assurant la satisfaction des 
C         conditions aux limites de Dirichlet;
C       Tri:reordonne les elements du tableau aa de 
C         maniere a ce que les elements d"une meme ligne
C         de la matrice A se trouvent ranges dans ce 
C         tableau dans l"ordre croissant de leurs 
C         indices de colonnes;
C       MCholeski:factorisation de Choleski de la matrice
C         A stockee sous forme Morse;
C       MSolvSys:resolution du systeme lineaire par 
C         descente-remontee; utilise la factorisee L de
C         Choleski de A stockee dans le tableau a sous
C         forme bande;
C       ecritF:ecriture de la solution approchee.
C       
C
C     fonctions:
C     ----------
C       errH10:calcul de l"erreur $H^1_0$ entre
C         la solution exacte phi0 et la solution 
C         approchee ff.
C        
C     bibliotheque (sous-programmes et fonctions):
C     -------------------------------------------
C       utilP1.for:fichier contenant les fonctions
C         deter (pour le calcul des aires), errH10
C         et les sous-programmes litT,litF, ecritF
C         et calBand;
C       SndMbre.for:fichier contenant les sous-
C         programmes calAire et SndMbre;
C       elem.for:fichier contenant la fonction
C         elemMat;
C       McalMat.for:fichier contenant les sous-
C         programmes McalMat et MDirCL;
C       MCholes.for:fichier contenant les sous-
C         programmes MCholeski et MSolvSys;
C       tri.for:fichier contenant le sous-
C         programme Tri.
C                  
C
      PARAMETER (nsMax=1800, ntMax=3400,nBandMax=70)            
      PARAMETER (nMLMax=(nsMax-1)*nBandMax+nsMax)           
      PARAMETER (nMAMax=8*nsMax)           
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
      DIMENSION aire(ntMax)		
      DIMENSION phi0(nsMax),f(nsMax),ff(nsMax)
      DIMENSION aa(nMAMax),ia(nsMax+1),ja(nMAMax)
      DIMENSION al(nMLMax),il(nsMax+1),jl(nMLMax)
      CHARACTER*20 nomfic			
C
      DATA cte/1.0e14/			
C				
      write(*,*) 'nom du fichier de stockage du maillage'
      read(*,'(a)') nomfic
      CALL litT(nomfic,ns,nt,q,ng,me,ngt)
      write(*,*) 'nom du fichier de stockage de f'
      write(*,*) '(second membre du Laplacien)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
      write(*,*) 'nom du fichier de stockage de phi0'
      write(*,*) '(conditions aux limites de Dirichlet)'
      read(*,'(a)') nomfic  
      CALL litF(nomfic,ns,phi0)
C
C
      CALL calBand(me,ng,nt,ns,nBand)
      CALL calAire(q,me,nt,ns,aire)
      CALL McalMat(q,me,ng,aire,nt,ns,aa,ia,ja)
      CALL SndMbre(me,f,ns,nt,aire,ff)
      CALL MDirCL(aa,ia,phi0,ns,ng,ff,cte)
C
      DO 1 i=1,ns
1     CALL Tri(ja(ia(i)),aa(ia(i)),ia(i+1)-ia(i))
C
      CALL MCholeski(ns,aa,ia,ja,al,il,jl,nBand)
      CALL MSolvSys(ns,ff,al,il,jl)
C
      err=errH10(q,me,ff,phi0,ns,nt)
      write(*,*) 'erreur=',err
      write(*,*) 'nom du fichier de stockage de la '
      write(*,*) 'solution approchee'
      read(*,'(a)') nomfic
      CALL ecritF(nomfic,ns,ff)  
      PAUSE
      END
C
C
C**********************************
C
C     BIBLIOTHEQUE
C
C**********************************
C
C
      INCLUDE 'elem.for'
      INCLUDE 'McalMat.for'
      INCLUDE 'MCholes.for'
      INCLUDE 'utilP1.for'
      INCLUDE 'SndMbre.for'
      INCLUDE 'tri.for'
C
C
