C Fichier: GCPLaplac.for
C   
      PROGRAM GCPLaplace
C
C Ce programme resout le probleme modele,i.e. le Laplacien
C avec conditions aux limites de Dirichlet :
C $-\Delta \Phi=f, \Phi$ donne sur le bord.
C
C Le probleme est ecrit sous forme variationnelle, puis 
C discretise par elements finis $P^1$. La solution appro-
C chee en chacun des sommets de la triangulation est  
C determinee en resolvant un systeme lineaire, systeme  
C qu'on resout par la methode du gradient conjugue 
C preconditionnee par $C=L^* (L^*)^T$ ou $(L^*)^T$ est 
C la factorisee de Choleski incomplete de $A$.
C Les matrices sont stockees sous forme Morse.
C
C
C     entree:
C     --------- 
C       triangulation (ns,nt,q,me,ng) et donnees du
C         probleme (second membre f de l'equation
C         de Laplace et condition a la limite de 
C         Dirichlet stockee dans le tableau phi0)
C         lues dans le fichier nomfic par l'intermediaire
C         des sous-programmes litT et litF;
C       parameter:nsMax (nombre maximal de sommets), 
C         ntMax (nombre maximal de triangles) de la
C         triangulation, nMAMax=8*nsMax (taille maximale
C         du tableau aa).
C                            
C     sortie:
C     ------- 
C       f:tableau contenant les valeurs de la solution 
C         approchee en chacun des sommets du maillage,
C         stocke dans un fichier (dont le nom est precise
C         par l"utilisateur) en vue d"une eventuelle
C         exploitation graphique.
C         Remarque: ce tableau contient au depart le 
C           second membre du Laplacien.
C                            
C     tableaux:
C     --------- 
C     maillage:
C       q(2,nsMax):coordonnees des sommets du maillage,
C       me(3,ntMax):numerotation des sommets des triangles,
C       ng(nsMax):tableau indiquant si un sommet est sur 
C          la frontiere ($ng(i)\ne 0$) ou non ($ng(i)=0$).
C
C     donnees:
C       f(nsMax): au depart, second membre du Laplacien,
C          (puis valeurs de la solution approchee),
C       phi0(nsMax): conditions aux limites de Dirichlet
C
C     tableaux pour le gradient conjugue preconditionne:
C       cg(nsMax):tableau contenant l'oppose du gradient
C         de la fonctionnelle a minimiser (produit de 
C         la matrice de preconditionnement C par g),
C       g(nsMax):tableau contenant g,
C       h(nsMax):tableau contenant la direction de 
C         descente,
C       
C
C     autres tableaux:
C       aa(nMAMax): tableau de stockage, sous forme Morse, 
C         de la matrice A,
C       ia(nsMax+1): indice de stockage dans le tableau aa
C         du premier element de la i eme ligne de A 
C         effectivement stocke dans aa,
C       ja(nMAMax): indice j de colonne de l"element 
C         Aij stocke;
C       al(nMLMax), ia(nsMax+1),ja(nMLMax): meme
C         signification que les 3 tableaux precedents,
C         mais pour la factorisee de Choleski incomplete
C         de A;
C       aire(ntMax):aire de chacun des triangles,
C       ff(nsMax):second membre du systeme lineaire a
C         resoudre;
C
C     scalaires:
C     ----------
C       cte: valeur de la tres grosse constante permettant
C         de satisfaire la condition a la limite de 
C         Dirichlet,
C       err: erreur en norme $H^1_0$ entre la solution 
C         exacte phi0 et la solution approchee ff.
C
C     fichiers:
C     ---------
C       nomfic:nom du fichier de stockage du maillage, 
C         des donnees du probleme modele (terme source du 
C         Laplacien et condition a la limite de Dirichlet),
C         ou de la solution approchee, en vue, par exemple,
C         d'une exploitation graphique.
C
C     sous-programmes:
C     ---------------- 
C       litT:lecture du maillage,
C       litF:lecture des donnees du probleme;
C       calAire:calcul de l"aire de chaque triangle;
C       McalMat:calcul et stockage, sous forme Morse dans
C         le tableau a, de la matrice A;
C       SndMbre:calcul, et stockage dans le tableau ff, 
C         du second membre du systeme;
C       MDirCL:programme assurant la satisfaction des 
C         conditions aux limites de Dirichlet;
C       Tri:reordonne les elements du tableau aa de 
C         maniere a ce que les elements d"une meme ligne
C         de la matrice A se trouvent ranges dans ce 
C         tableau dans l"ordre croissant de leurs 
C         indices de colonnes;
C       ICholeski:factorisation de Choleski incomplete 
C         de la matrice A stockee sous forme Morse;
C       GradC:programmation du gradient conjugue 
C         preconditionne;
C         N.B.: ce sous-programme necessite les
C         sous-programmes IAmul et ISolvSys;
C       ecritF:ecriture de la solution approchee.
C       
C
C     fonctions:
C     ----------
C       errH10:calcul de l"erreur $H^1_0$ entre
C         la solution exacte phi0 et la solution 
C         approchee f.
C        
C     bibliotheque (sous-programmes et fonctions):
C     -------------------------------------------
C       utilP1.for:fichier contenant les fonctions
C         deter (pour le calcul des aires), errH10
C         et les sous-programmes litT,litF, ecritF
C         et calBand;
C       SndMbre.for:fichier contenant les sous-
C         programmes calAire et SndMbre;
C       elem.for:fichier contenant la fonction
C         elemMat;
C       McalMat.for:fichier contenant les sous-
C         programmes McalMat et MDirCL;
C       ICholes.for:fichier contenant le sous-
C         programme ICholeski;
C       GradCP.for:fichier contenant le sous-
C         programme GradC;
C       tri.for:fichier contenant le sous-
C         programme Tri.
C                  
C
      PARAMETER (nsMax=1800, ntMax=3400,nMAMax=8*nsMax)            
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
      DIMENSION aire(ntMax)		
      DIMENSION phi0(nsMax),f(nsMax),ff(nsMax)
      DIMENSION aa(nMAMax),ia(nsMax+1),ja(nMAMax)
      DIMENSION al(nMAMax)
      DIMENSION g(nsMax),cg(nsMax),h(nsMax)

      CHARACTER*20 nomfic			
C
      DATA cte/1.0e25/			
      DATA eps/1.0e-12/			
      DATA kMax/150/			
C				
      write(*,*) 'nom du fichier de stockage du maillage'
      read(*,'(a)') nomfic
      CALL litT(nomfic,ns,nt,q,ng,me,ngt)
      write(*,*) 'nom du fichier de stockage de f'
      write(*,*) '(second membre du Laplacien)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
      write(*,*) 'nom du fichier de stockage de phi0'
      write(*,*) '(conditions aux limites de Dirichlet)'
      read(*,'(a)') nomfic  
      CALL litF(nomfic,ns,phi0)
C
C
      CALL calAire(q,me,nt,ns,aire)
      CALL McalMat(q,me,ng,aire,nt,ns,aa,ia,ja)
      CALL SndMbre(me,f,ns,nt,aire,ff)
      CALL MDirCL(aa,ia,phi0,ns,ng,ff,cte)
C
      DO 1 i=1,ns
1     CALL Tri(ja(ia(i)),aa(ia(i)),ia(i+1)-ia(i))
C
      CALL ICholeski(ns,aa,ia,ja,al)
      DO 2 i=1,ns
        f(i)=0.
        if(ng(i).NE.0) f(i)=phi0(i)
2     CONTINUE
      CALL GradC(ns,aa,ia,ja,al,ff,f,g,h,cg,kMax,eps)
C
      err=errH10(q,me,f,phi0,ns,nt)
      write(*,*) 'erreur=',err
      write(*,*) 'nom du fichier de stockage de la '
      write(*,*) 'solution approchee'
      read(*,'(a)') nomfic
      CALL ecritF(nomfic,ns,f)  
      PAUSE
      END
C
C
C-----------------------------------------------
      SUBROUTINE IAmul(ns,aa,ia,ja,x,b)
C-----------------------------------------------
C Ce sous-programme calcule le produit de la matrice
C A, stockee sous forme Morse dans les tableaux 
C (aa,ia,ja), par le vecteur x; le resultat est dans le
C tableau b.
C Le calcul se fait d"abord en calculant la contribution
C associee a la partie triangulaire inferieure de la 
C matrice (qui est celle stockee dans le tableau aa),
C puis celle associee a la partie triangulaire superieure
C stricte.
C
C
      DIMENSION x(ns), b(ns)
      DIMENSION aa(*),ia(ns),ja(*)
      DO 1 i=1,ns
        somme=0.
        DO 2 k=ia(i),ia(i+1)-1
          somme=somme+aa(k)*x(ja(k))
2       CONTINUE
        b(i)=somme
1     CONTINUE
      DO 3 i=1,ns
        xi=x(i)
        DO 3 k=ia(i),ia(i+1)-2
          jak=ja(k)
          b(jak)=b(jak)+aa(k)*xi
3     CONTINUE
      END
C
C
C-----------------------------------------------
            SUBROUTINE ISolvSys(ns,z,al,il,jl)
C-----------------------------------------------
C Ce sous-programme resout le systeme lineaire
C $LL^T z = r$, ou L est une matrice triangulaire
C inferieure de taille ns x ns, stockee sous forme
C Morse dans les tableaux (al,il,jl). En entree, le 
C tableau z contient le second membre du systeme; 
C en sortie, il contient la solution z.
C La resolution se fait en 2 etapes: une etape de 
C descente, suivie d'une etape de remontee.
C
C 
      DIMENSION z(*),al(*),il(*),jl(*)
C
      DO 1 i=1,ns
        somme=z(i)
        DO 2 k=il(i), il(i+1)-2
          somme=somme-al(k)*z(jl(k))
2       CONTINUE
        z(i)=somme/al(il(i+1)-1)
1     CONTINUE
C
      DO 3 i=1,ns
        z(i) = z(i)/al(il(i+1)-1)
3     CONTINUE
      DO 4 i1=1,ns
        i=ns+1-i1
        DO 4 k=il(i), il(i+1)-2
          jlk=jl(k)
          z(jlk) = z(jlk) - al(k)*z(i)/al(il(jlk+1)-1)
4       CONTINUE
      END
C
C
C
C
C**********************************
C
C     BIBLIOTHEQUE
C
C**********************************
C
C
      INCLUDE 'elem.for'
      INCLUDE 'GradCP.for'
      INCLUDE 'McalMat.for'
      INCLUDE 'ICholes.for'
      INCLUDE 'utilP1.for'
      INCLUDE 'SndMbre.for'
      INCLUDE 'tri.for'
