C Fichier: gEDP.for
C
   
      PROGRAM gEDP
C
C Ce programme resout le probleme elliptique du second 
C ordre general suivant ($\Gamma=\Gamma_1 \cup \Gamma_2$):
C       $ a_{0}\varphi  -\nabla .(M\nabla \varphi)  =f$ dans $\Omega,$
C       $  \varphi =\varphi _{0} $ sur $\Gamma _{1},$   
C       $a_{1}\varphi + n^{T}M\nabla \varphi =g $ sur $ \Gamma _{2},$
C ecrit sous forme variationnelle, puis discretise par 
C elements finis $P^1$. La solution approchee en chacun
C des sommets de la triangulation verifie le systeme 
C lineaire $A \Phi=F$, systeme qu"on resout par une 
C methode de gradient conjugue preconditionne par la
C diagonale de $A$. La matrice $A$ de ce systeme est stockee
C sous forme Morse.
C
C
C     entree:
C     --------- 
C       triangulation (ns,nt,q,me,ng) lue dans le fichier
C         nomfic par l'intermediaire du sous-programme
C         litT;
C       donnees du probleme lues dans le fichier nomfic, 
C         dont le nom est a chaque fois precise par 
C         l"utilisateur au moment de l"execution du 
C         programme, grace au sous-programme litF; les 
C         donnees sont lues dans l"ordre suivant:
C         - condition a la limite de Dirichlet lue dans le
C             sous-programme lecDon et stockee (si elle 
C             existe effectivement) dans le tableau phi0;
C             a ce tableau est associe un parametre 
C             indicateur note iDir, qui vaut 0 s"il n"y a
C             pas de condition aux limites de Dirichlet, 
C             et 1 sinon;
C         - matrice $M$ et fonction $a_0$ de l"EDP lues
C             dans le sous-programme lecDon, ainsi que
C             les fonctions $a_1$ et $g$ de la condition a la 
C             limite de Robin; le tableau AM contient les 
C             valeurs moyennes sur chaque triangle des 
C             coefficients de la matrice $M$, le tableau a0 
C             contient les valeurs en chacun des sommets du  
C             maillage de la fonction $a_0$, et il en est de 
C             meme pour le tableau a1 associe a la fonction 
C             $a_1$, ainsi que pour le tableau f ou est 
C             stockee la fonction $g$.
C             Aux 3 tableaux AM,a0,a1 sont associes des  
C             parametres indicateurs respectivement notes 
C             iM,ia0,ia1 (cf sous-programme lecDon pour les
C             details);
C         - second membre $f$ de l'equation de Laplace lu
C             dans le sous-programme gSMbre et stocke
C             dans le tableau f (apres avoir calcule 
C             l"integrale de bord associee a la fonction 
C             $g$).
C       parameter:nsMax (nombre maximal de sommets), 
C         ntMax (nombre maximal de triangles) de la
C         triangulation, nMAMax=8*nsMax (taille maximale
C         du tableau aa).
C                            
C     sortie:
C     ------- 
C       f:tableau contenant les valeurs de la solution 
C         approchee en chacun des sommets du maillage,
C         stocke dans un fichier (dont le nom est precise
C         par l"utilisateur) en vue d"une eventuelle
C         exploitation graphique.
C       Remarque: ce tableau est utilise tout au long
C           du programme comme tableau auxiliaire (avant
C           de contenir les valeurs de la solution).
C                            
C     tableaux:
C     --------- 
C     maillage:
C       q(2,nsMax):coordonnees des sommets du maillage,
C       me(3,ntMax):numerotation des sommets des
C         triangles,
C       ng(nsMax):tableau indiquant si un sommet est sur 
C          la frontiere ($ng(i) \ne 0$) ou s"il est 
C          interne ($ng(i)=0$).
C          
C
C     donnees:
C       f(nsMax): au depart, coefficients de la matrice
C           $M$, puis fonction $g$, enfin second membre 
C            $f$ du laplacien,
C          (a la fin: valeurs de la solution approchee);
C       phi0(nsMax): conditions aux limites de Dirichlet;
C         N.B.: ce tableau phi0 permet de caracteriser
C               les sommets $q^i$ de $\Gamma_1$ par la 
C               relation phi0(i) $\ne$ 0; ceci signifie
C               qu"en amont de ce programme, il est 
C               necessaire de modifier la valeur de ce  
C               tableau en des points de Dirichlet $q^i$ 
C               pour lesquels on aurait: $\varphi_0(q^i)=0$
C               (les remplacer par une tr\`es petite 
C               constante).
C       AM(3*ntMax): valeurs moyennes des coefficients de
C         la matrice $M$ sur chaque triangle; les nt 
C         premiers indices correspondent au stockage des
C         coefficients d"indices (1,1), les nt suivants
C         ceux d"indices (2,2), et enfin les derniers, 
C         ceux d"indices (1,2), selon les valeurs possibles
C         du parametre indicateur iM (cf sous-programme
C         lecDon);
C       a0(nsMax),a1(nsMax): valeurs des fonctions $a_0$ et
C         $a_1$ en chacun des sommets du maillage, avec 
C         pour parametres indicateurs ia0 et ia1.
C         (le tableau a0 sert ensuite comme tableau 
C          auxiliaire dans le gradient conjugue )
C
C     tableaux pour le gradient conjugue preconditionne:
C       cg(nsMax):tableau contenant l'oppose du gradient
C         de la fonctionnelle a minimiser (produit de 
C         la matrice de preconditionnement C par g),
C       g(nsMax):tableau contenant g,
C       h(nsMax):tableau contenant la direction de 
C         descente,
C       a0(nsMax): tableau contenant l"inverse des
C         elements diagonaux de la matrice.
C       N.B. dans le cas du preconditionnement par la
C         diagonale choisi, nous aurions pu nous passer
C         du tableau cg; nous l"avons garde ici pour ne
C         pas trop modifier le sous-programme du gradient
C         conjugue par rapport au chapitre precedent.
C       
C
C     autres tableaux:
C       aa(nMAMax): tableau de stockage, sous forme Morse, 
C         de la matrice A,
C       ia(nsMax+1): indice de stockage dans le tableau aa
C         du premier element de la i eme ligne de A 
C         effectivement stocke dans aa,
C       ja(nMAMax): indice j de colonne de l"element 
C         Aij stocke;
C       aire(ntMax):aire de chacun des triangles,
C       ff(nsMax):second membre du systeme lineaire a
C         resoudre;
C
C     scalaires:
C     ----------
C       cte: valeur de la tres grosse constante permettant
C         de satisfaire la condition a la limite de 
C         Dirichlet,
C       err: erreur en norme $H^1_0$ entre la solution 
C         exacte phi0 et la solution approchee ff.
C
C     fichiers:
C     ---------
C       nomfic:nom du fichier de stockage du maillage, 
C         des donnees du probleme modele, ou de la solution
C         approchee, en vue, par exemple, d'une 
C         exploitation graphique.
C
C     sous-programmes:
C     ---------------- 
C       litT:lecture du maillage,
C       calAire:calcul de l"aire de chaque triangle;
C       lecDon:lecture des donnees du probleme: matrice $M$ 
C         de l"operateur aux derivees partielles, fonctions 
C         $a_1$ et $g$ de la condition a la limite de Robin;
C       gDirCL:programme assurant la satisfaction des 
C         conditions aux limites de Dirichlet;
C       gSMbre:calcul, et stockage dans le tableau ff, 
C         du second membre du systeme;
C       gcalMat:calcul et stockage, sous forme Morse dans
C         le tableau aa, de la matrice A;
C       Tri:reordonne les elements du tableau aa de 
C         maniere a ce que les elements d"une meme ligne
C         de la matrice A se trouvent ranges dans ce 
C         tableau dans l"ordre croissant de leurs 
C         indices de colonnes;
C       gGradC:programmation du gradient conjugue 
C         preconditionne;
C       ecritF:ecriture de la solution approchee.
C       
C
C     fonctions:
C     ----------
C       errH10:calcul de l"erreur $H^1_0$ entre
C         la solution exacte phi0 et la solution 
C         approchee f.
C        
C     bibliotheque (sous-programmes et fonctions):
C     -------------------------------------------
C       utilP1.for:fichier contenant les fonctions
C         deter (pour le calcul des aires), errH10
C         et les sous-programmes litT,litF et ecritF;
C       SndMbre.for:fichier contenant les sous-
C         programmes calAire;
C       gelem.for:fichier contenant la fonction
C         geMat et les sous-programmes gDirCL et
C         gSMbre;
C       tri.for:fichier contenant le sous-
C         programme Tri;
C                  
C
      PARAMETER (nsMax=1800, ntMax=3400,nMAMax=8*nsMax)            
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
      DIMENSION aire(ntMax)		
      DIMENSION AM(3*ntMax),a0(nsMax),a1(nsMax)
      DIMENSION phi0(nsMax),f(nsMax),ff(nsMax)
      DIMENSION aa(nMAMax),ia(nsMax+1),ja(nMAMax)
      DIMENSION g(nsMax),cg(nsMax),h(nsMax)

      CHARACTER*20 nomfic			
C
      DATA cte/1.0e25/			
      DATA eps/1.0e-12/			
      DATA kMax/150/			
C				
      write(*,*) 'nom du fichier de stockage du maillage'
      read(*,'(a)') nomfic
      CALL litT(nomfic,ns,nt,q,ng,me,ngt)
      CALL calAire(q,me,nt,ns,aire)
      CALL lecDon(q,me,ng,ns,nt,AM,iM,a0,ia0,a1,ia1,f
     >            ,phi0,iDir)
      CALL gSMbre(me,f,ns,nt,aire,ff,ng,q)
      CALL gcalMat(q,me,ng,aire,ns,nt,AM,iM,a0,ia0,a1,ia1,
     >             aa,ia,ja)
      IF(iDir.EQ.1) THEN
        CALL gDirCL(aa,ia,phi0,ng,ns,cte)
      ENDIF
C
C
      DO 1 i=1,ns
1     CALL Tri(ja(ia(i)),aa(ia(i)),ia(i+1)-ia(i))
C
      DO 2 i=1,ns
        f(i)=0.
        if(ng(i).NE.0) f(i)=phi0(i)
2     CONTINUE
      CALL gGradC(ns,aa,ia,ja,a0,ff,f,g,h,cg,kMax,eps)
C
      IF(iDir.EQ.1) THEN
        err=errH10(q,me,f,phi0,ns,nt)
        write(*,*) 'erreur=',err
      ENDIF
C
      write(*,*) 'nom du fichier de stockage de la '
      write(*,*) 'solution approchee'
      read(*,'(a)') nomfic
      CALL ecritF(nomfic,ns,f)  
      PAUSE
      END
C
C
C---------------------------------------------------------
      SUBROUTINE lecDon(q,me,ng,ns,nt,AM,iM,a0,ia0,a1,
     >                  ia1,f,phi0,iDir)
C---------------------------------------------------------
C
C Ce sous-programme lit (dans des fichiers dont le nom est
C precise par l"utilisateur au moment de l"execution) 
C et stocke certaines donnees du probleme, dans l"ordre
C suivant:
C
C - il demande a l"utilisateur s"il y a des conditions
C   aux limites de type Dirichlet (iDir=1) ou non
C   (iDir=0); dans le premier cas, il lit dans un 
C   fichier, dont le nom est precise par l"utilisateur au
C   moment de l"execution du programme, les valeurs de 
C   phi0 et les stocke dans le tableau phi0.
C   N.B. Nous rappellons qu"il est necessaire pour la 
C        suite du programme que ce tableau ne contienne
C        aucune valeur nulle.
C 
C - il demande a l"utilisateur si l"EDP a un terme constant
C   $a_0$ (ia0=1) ou non (ia0=0); dans le premier cas, il
C   lit cette fonction et stocke ses valeurs en chacun des 
C   sommets du maillage dans le tableau a0;
C
C - il demande a l"utilisateur si le coefficient $a_1$ de
C   la condition a la limite de Robin est nul (ia1=0) ou
C   non (ia1=1); dans le dernier cas, il lit cette
C   fonctionet stocke ses valeurs en chacun des sommets
C   du maillage dans le tableau a1, puis annule les 
C   valeurs de ce tableau qui correspondent a des points
C   du bord $\Gamma_1$.
C
C - il demande a l"utilisateur quel est le type de l"EDP
C   etudiee:
C   * s"il s"agit du Laplacien ($M=Id$), le parametre 
C     indicateur iM vaut 0 et on ne fait rien;
C   * si la matrice $M=M_{11} Id$, mais differente de
C     la matrice identite, iM vaut 1, on lit les valeurs
C     de la fonction $M_{11}$ en chacun des sommets du 
C     maillage, puis on calcule la valeur moyenne de cette
C     fonction sur chacun des triangles (i.e. le tiers de  
C     la somme des valeurs prises en chacun des sommets du
C     triangle); le resultat est stocke dans les nt 
C     premiers indices du tableau AM;
C   * si la matrice est diagonale, on lit les valeurs de
C     la fonction $M_{22}$ en chacun des sommets du maillage
C     et on stocke la valeur moyenne sur chaque triangle
C     dans le tableau AM a la suite des valeurs 
C     precedentes; si de plus la matrice n"est 
C     pas du type precedent, le parametre iM vaut 2;
C   * dans le cas general, on lit aussi les valeurs de $M_{12}$
C     et on stocke leur valeur moyenne sur chaque triangle, 
C     toujours dans le tableau AM entre les indices 2*nt+1
C     et 3*nt de ce tableau; si de plus la matrice n"est 
C     pas du type precedent, la parametre iM vaut 3.
C
C - enfin il lit et stocke dans le tableau f les valeurs de
C   la fonction $g$ qui intervient dans les conditions aux 
C   limites de Robin, en prolongeant celle-ci par $0$ pour
C   les points du bord $\Gamma_1$.
C
C
      PARAMETER (nsMax=1800, ntMax=3400,nMAMax=8*nsMax)            
      DIMENSION q(2,nsMax),me(3,ntMax),ng(nsMax)
      DIMENSION AM(3*ntMax),a0(nsMax),a1(nsMax),phi0(nsMax)
      DIMENSION f(nsMax)

      CHARACTER*20 nomfic			
C
      write(*,*) 'y a t il des conditions de Dirichlet?'
      write(*,*) '(reponse: 1 = oui, 0 = non)'
      read(*,*) iDir
      IF (iDir.EQ.0) THEN
        write(*,*) 'nom du fichier de stockage de phi0'
        write(*,*) '(condition a la limite de Dirichlet)'
        read(*,'(a)') nomfic
        CALL litF(nomfic,ns,phi0)
      ENDIF
C
      write(*,*) 'y a t" il un terme constant dans l"EDP?'
      write(*,*) '(reponse: 1 = oui, 0 = non)'
      read(*,*) ia0
      IF (ia0.EQ.0) GO TO 1
      write(*,*) 'nom du fichier de stockage de a0'
      write(*,*) '(coefficient constant de l"EDP)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,a0)
C
C
1     write(*,*) 'y a t"il un coeff a1 dans la C.L.?'
      write(*,*) '(reponse: 1 = oui, 0 = non)'
      read(*,*) ia1
      IF (ia1.EQ.0) GO TO 2
      write(*,*) 'nom du fichier de stockage de a1'
      write(*,*) '(coefficient a1 de la condition de 
     >              Robin)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,a1)
C
      IF(iDir.EQ.1) THEN
        DO 5 i=1,ns
          IF((ng(i).NE.0).AND.(phi0(i).NE.0)) THEN
            a1(i)=0.
          ENDIF
5       CONTINUE
      ENDIF
C
C
C
2     write(*,*) 'description de la matrice M de l"EDP'
      write(*,*) 'entrer la valeur du parametre iM'
      write(*,*) 'rappel: iM=0 pour M=Id (Laplacien)'
      write(*,*) 'iM=1 pour M=M11*Id et M11 different de 1'
      write(*,*) 'iM=2 pour M diagonale et non du type 
     >              precedent'
      write(*,*) 'iM=3 sinon'
      read(*,*) iM
C
      nkt=0
C
      IF (IM.EQ.0) GO TO 100
        write(*,*) 'nom du fichier de stockage de M11'
        read(*,'(a)') nomfic
        CALL litF(nomfic,ns,f)
        DO 10 k=1,nt
          kt=nkt+k
          som=0.
          DO 11 il=1,3
            i=me(il,k)
            som=som+f(i)
11        CONTINUE
          AM(kt)=som/3.
10      CONTINUE
      nkt=nkt+nt
C
      IF (IM.EQ.1) GO TO 100
        write(*,*) 'nom du fichier de stockage de M22'
        read(*,'(a)') nomfic
        CALL litF(nomfic,ns,f)
        DO 20 k=1,nt
          kt=nkt+k
          som=0.
          DO 21 il=1,3
            i=me(il,k)
            som=som+f(i)
21        CONTINUE
          AM(kt)=som/3.
20      CONTINUE
      nkt=nkt+nt
C
      IF (IM.EQ.2) GO TO 100
        write(*,*) 'nom du fichier de stockage de M12'
        read(*,'(a)') nomfic
        CALL litF(nomfic,ns,f)
        DO 30 k=1,nt
          kt=nkt+k
          som=0.
          DO 31 il=1,3
            i=me(il,k)
            som=som+f(i)
31        CONTINUE
          AM(kt)=som/3.
30      CONTINUE
      nkt=nkt+nt
C
100   CONTINUE
C
C
      write(*,*) 'nom du fichier de stockage de g'
      write(*,*) '(condition a la limite de Robin)'
      read(*,'(a)') nomfic
      CALL litF(nomfic,ns,f)
C
      IF(iDir.EQ.1) THEN
        DO 40 i=1,ns
          IF((ng(i).NE.0).AND.(phi0(i).NE.0)) THEN
            f(i)=0.
          ENDIF
40      CONTINUE
      ENDIF
C
C
      END
C
C
C---------------------------------------------------------
      SUBROUTINE gcalMat(q,me,ng,aire,ns,nt,AM,iM,a0,ia0,
     >                    a1,ia1,aa,ia,ja)
C---------------------------------------------------------
C Construit la partie triangulaire inferieure de la matrice
C du systeme lineaire a resoudre et stocke le resultat dans
C le tableau aa sous forme Morse, avec les tableaux 
C auxiliaires d"indices ia et ja: si l"element $A_{ij}$ est 
C stocke dans le tableau aa a un indice k, alors ja(k)=j. 
C ia(i) est l"indice de stockage dans le tableau aa du 
C premier coefficient de la i eme ligne de $A$ effectivement
C stocke dans ce tableau.
C
C En convenant de poser ia(ns+1)=nElemA+1, 
C l"element Aii de la matrice A est stocke dans 
C le tableau aa a l"indice ia(i+1)-1.
C Au debut du calcul, le tableau ia a une autre
C signification: ia(i) est le nombre de sommets
C strictement voisins du point d"indice i
C (i.e. exception faite du point lui-meme).
C
C
C     entree:
C     ------
C     triangulation (ns,nt,q,me,ng) et les aires des
C       triangles stockees dans le tableau aire.
C     donnees du probleme (AM,iM,a0,ia0,a1,ia1)
C
C     sortie:
C     ------
C     tableaux aa,ia,ja; taille nElemA du tableau aa
C
C    
C     remarques importantes:
C     ---------------------
C     - on suppose les triangles orientes positivement
C     - le tableau ng sert a detecter les aretes 
C       frontieres de la maniere suivante: une arete
C       est frontiere si et seulement si ses 2 
C       extremites sont des sommets frontieres.
C
C     Il faut donc verifier que la triangulation 
C     etudiee satisfait ces deux conditions.
C
C
      DIMENSION q(2,ns),me(3,nt),ng(ns),aire(nt)
      DIMENSION aa(*),ia(*),ja(*)
C
      DO 110 i=1,ns
110     ia(i)=0
      DO 111 k = 1,nt
        DO 111 il = 1,3
          ilp = mod(il,3)+1
          i=me(il,k)
          ip=me(ilp,k)
          IF (ip.LE.i) THEN
             ia(i) = ia(i)+1
          ELSEIF ((ng(i).NE.0).AND.(ng(ip).NE.0)) THEN 
             ia(ip) = ia(ip)+1
          ENDIF	
111     CONTINUE
C
C ia(i) est le nombre de sommets voisins stricts 
C du sommet d"indice i.
C
      nElemA=0
      DO 211 i = 1,ns
        ivois=ia(i)
        ia(i)= nElemA + 1
        nElemA = nElemA + ivois + 1
211   CONTINUE
      ia(ns+1)=nElemA+1
C
C ia est construit; reste a construire aa et ja.
C
      DO 1 k = 1, nElemA
      aa(k)=0.
1     ja(k)=0
      DO 2 kt = 1,nt
        DO 2 il=1,3
          i = me(il,kt)
          DO 2 jl=1,3
            j=me(jl,kt)
            k=0
            IF (i.EQ.j) THEN
              k=ia(i+1)-1
              ja(k)=i
            ELSEIF (i.GT.j) THEN
              DO 3 l= ia(i), ia(i+1)-2
C
C j ne peut pas etre ailleurs
C
                IF (ja(l).EQ.0) GOTO 4
                IF ( ja(l) .EQ. j )THEN
                  k=l
                  GOTO 4
                ENDIF
3             CONTINUE
              write(*,*)'erreur dans McalMat.for au point
     >                     i,j=',i,j
4             IF (k.EQ.0) THEN 
                k = l
                ja(l) = j 
              ENDIF
            ENDIF
            IF(k.NE.0)aa(k)=aa(k)+
     >         geMat(kt,il,jl,aire(kt),q,me,ng,ns,nt,
     >              AM,iM,a0,ia0,a1,ia1)
2     CONTINUE
      END
C
C
C----------------------------------------------------------
      SUBROUTINE gGradC(n,aa,ia,ja,a0,b,x,g,h,cg,kMax,eps)
C----------------------------------------------------------
C Ce sous-programme resout le systeme lineaire $Ax=b$ par
C gradient conjugue preconditionne, avec au plus kMax 
C iterations. Le test d"arret est: $|Ax^k-b|<eps|Ax^0-b|$.
C La matrice $A$ du systeme lineaire, de taille $n \times n$
C est stockee sous forme Morse dans le tableau aa;
C ce stockage necessite aussi les tableaux auxiliaires
C d"indices ia et ja.
C La matrice $C$ de preconditionnement est la diagonale de
C $A$ (les inverses des valeurs de cette diagonale sont 
C stockees dans le tableau a0).
C
C     entree:
C     --------
C       matrice $A$ du systeme stockee sous forme Morse 
C         dans les tableaux (aa,ia,ja),
C       second membre de l"equation stocke dans le 
C         tableau b;
C       nombre maximal kMax d"iterations, precision eps
C         du calcul;
C
C     N.B.: le calcul du second membre ne servant qu"a 
C       l"etape d"initialisation, le tableau b est ensuite
C       ecrase et utilise comme tableau auxiliaire de  
C       calcul, en particulier pour le stockage du produit 
C       de la matrice $A$ par $h$, puis pour le stockage 
C       de la solution y du systeme $C y= A h$.
C       
C                            
C     sortie:
C     ------- 
C       x:tableau contenant les valeurs de la solution 
C         en chacun des sommets du maillage;
C       gcg:scalaire contenant le carre de la norme de
C         $g$ (en pratique ce nombre doit decroitre au 
C         cours des iterations).
C                            
C     autres tableaux:
C     ----------------
C       g, cg: tableaux contenant les valeurs de $g^n$ et
C         de $C g^n$ respectivement;
C       h: tableau contenant la direction de descente $h^n$
C         (au tout debut du programme, ce tableau sert a 
C          stocker le produit de $A$ par $x^0$);
C       
C
C     sous-programmes:
C     ---------------- 
C       IAmul:effectue le produit d"une matrice par 
C         un vecteur, la matrice etant stockee sous
C         forme Morse,
C       Ce sous-programme est dans le fichier
C       GCPLaplac.for.
C
C
      DIMENSION x(*),b(*),g(*),h(*),cg(*)
      DIMENSION aa(*),ja(*),ia(*),a0(*)
C
      DO 20 i=1,n
      a0(i)=1./aa(ia(i+1)-1)
20    CONTINUE
C
      DO 1 k=1, kMax
C
        IF (k.EQ.1) THEN
          CALL IAmul(n,aa,ia,ja,x,h)
          DO 2 i=1,n
            cg(i)=b(i)-h(i)
            g(i)=cg(i)*a0(i)
2         CONTINUE
          gcg=0.
          DO 3 i=1,n
            gcg=gcg + g(i)*cg(i)
            h(i)=g(i)
3         CONTINUE
        ENDIF
C
        CALL IAmul(n,aa,ia,ja,h,b)
        hAh=0.
        DO 10 i=1,n		
          hAh=hAh+h(i)*b(i)
10      CONTINUE
        ro=gcg/hAh
        DO 11 i=1,n
          cg(i) = cg(i) -ro*b(i)
          b(i)=b(i)*a0(i)
11      CONTINUE
        DO 12 i=1,n
          x(i)=x(i)+ro*h(i)
          g(i)=g(i)- ro*b(i)
12      CONTINUE
        gamma=gcg
        gcg=0
        DO 13 i=1,n
          gcg=gcg + g(i)*cg(i)
13      CONTINUE
C
        write(*,*) 'iteration:',k,'gcg=',gcg
        IF(k.EQ.1) eps=eps*gcg
        IF (gcg.LT.eps) RETURN
        gamma=gcg/gamma
        DO 14 i=1,n
          h(i)=g(i)+gamma*h(i)
14      CONTINUE
C
1     CONTINUE
      END
C
C
C-----------------------------------------------
      SUBROUTINE IAmul(ns,aa,ia,ja,x,b)
C-----------------------------------------------
C Ce sous-programme calcule le produit de la matrice
C A, stockee sous forme Morse dans les tableaux (aa,ia,ja),
C par le vecteur x; le resultat est dans le tableau b.
C
C
      DIMENSION x(ns), b(ns)
      DIMENSION aa(*),ia(ns),ja(*)
      DO 1 i=1,ns
        somme=0.
        DO 2 k=ia(i),ia(i+1)-1
          somme=somme+aa(k)*x(ja(k))
2       CONTINUE
        b(i)=somme
1     CONTINUE
      DO 3 i=1,ns
        xi=x(i)
        DO 3 k=ia(i),ia(i+1)-2
          jak=ja(k)
          b(jak)=b(jak)+aa(k)*xi
3     CONTINUE
      END
C
C
C**********************************
C
C     BIBLIOTHEQUE
C
C**********************************
C
C
      INCLUDE 'gelem.for'
      INCLUDE 'utilP1.for'
      INCLUDE 'SndMbre.for'
      INCLUDE 'tri.for'
C
C
